##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: analyse.R, v22
# Collection of generic code snipets for analysing results, plotting etc. mostly using utilities.R
# Not called by anything. File is not to be sourced.

##### Plot gamma function

# The gamma function is shape invariant under multiplication by a common factor of its parameters (or of mean and sd).
# That is, if we multiply the parameters and the size of the x-scale by a factor, the plot looks unchanged.

plot.gamma(scale=5, shape=1, yzero=TRUE, xlim=c(0,15))

# plot gamma distribution
# mean is shape*scale, variance is shape*(scale^2)
# => scale=var/mean, shape=mean^2/var
# mode is (shape-1)*scale (for shape>1)
pdf("gamma.pdf", height=10, width=14, paper="a4r")
par(mfrow=c(6,6),mar=c(3,3,1,1), oma=c(0.8,0.8,1.4,0.8), mgp=c(1.5,0.5,0))
x <- seq(from=0,to=10,by=0.1)
for(mean in c(0.5, 1:5)) {
  for(var in c(0.5, 1:5)^2) {
    y <- dgamma(x, scale=var/mean, shape=mean^2/var)
    plot(x, y, xlab="x", ylab="dgamma(x)", yaxs="i", type="l", col="red", main=paste(mean, sqrt(var)) )
    lines(x, dnorm(x, mean, sqrt(var)), yaxs="i", col="blue")
  }
} 
dev.off()


##### Plot data, generate/plot simulated data

plot.data(obsdata=read.obs.dat("~/Desktop/quasarlightcurves/lc_22.5595.1333.R.mjd", obsdataHeader=FALSE))

plot.data(obsdata=sim.const(simseed=100), fname=NA, yzero=FALSE)
plot.data(obsdata=sim.asfuncquasisinusoid(simseed=100), fname=NA, yzero=FALSE)
plot.data(obsdata=sim.OUprocess(simseed=301), fname="tsdata.pdf", yzero=FALSE)
plot.data(obsdata, fname="tsdata.pdf", yzero=FALSE)

source("ctsmod_genfunctions.R")
oplot.model.data(obsdata=sim.sinusoid(simseed=102), TSMod1class="FuncSinusoidZero", paramlist=list(sinamp=0.02, freq=0.10, phase=0, sd=0.01), fname="obsdata.pdf")


##### Summarize all results

sumres <- function() {
  cat("logLike(No-model) =", logLikeNomod, "\n")
  if(exists("logEvidence")) {
    cat("log(Evidence) =", logEvidence, "\n")
    temp <- log10(mean(10^(evCalc[,2])))
    cat("log(Evidence) =", temp, " (with hfac=0)\n")
  }
  if(exists("logLike")) {
    cat("log(posterior-averaged likelihood) =", logLike, "\n")
#    cat("DIC=D+2Neff =", dic$dic,  "   D =", dic$dic-2*dic$effNpar, "   Neff =", dic$effNpar, "\n-(D/2)/ln10 =", -1/2*(dic$dic-2*dic$effNpar)/log(10), "(=logLike[averageTheta])\n")
  }
  if(exists("logPTEv")) {
    cat("Thermodynamic integration approximation of log10(evidence) =", logPTEv, "\n")
  }
  if(exists("kfoldCV")) {
    cat("kfoldCV log(likelihood) =", sum(kfoldCV$logLikePart), "\n")
  }
}
sumres()


##### Plot/analyse evidence calculation results

# If using R2Cuba then cuhre gives lots of blank lines in the output file. Can remove and rewrite using:
write(file="output.mod", noquote(drop(t(read.delim("output", sep="\n", header=FALSE, blank.lines.skip=TRUE)))))

# Misc
load("test.Robj")
source("utilities.R")
t(colnames(evCalc)) # to show pcol values to use
range(evCalc[,2])   # check whether we need to use hfac
length(which(evCalc[,2]==-Inf))
system("wc test.warnings")

# Calculate statistics from posterior
calc.post.stats(evCalc is.ev=TRUE) # From v19. Not yet tested
calc.xic(evCalc, thetaFlag, obsdata, logEvNomod)

# Plot PDFs from posterior
hfac <- 0
plot.1dpost(evCalc, hfac=hfac, pcol=c(3,4), plotsize=c(6,4), mfrow=c(1,2))
plot.2dpost(evCalc, hfac=hfac, dynrange=10, pcol=c(3,4), fname="posterior2D_relax_diffcon.pdf")


##### Overplot model on data

oplot.model.data(obsdata, TSMod1class="FuncCombUniformSigmoidZero", fname="lcurve_FuncCombUniformSigmoidZero.pdf", paramlist=as.list(find.peak.sol(postSamp)))
oplot.model.data(obsdata, TSMod1class="FuncLinear", fname="lcurve_FuncLinear.pdf", paramlist=list(priorform="Gaussian", slope=0.920, tmid=47.05, offset=-2.726))
oplot.model.data(obsdata, TSMod1class="FuncCombUniformQuasiSinusoidZero", fname="lcurve_TrueModel.pdf", paramlist=list(priorform="Gaussian", offset=0, sinamp=316, freq=1/62, phase=5.2/(2*pi), fp1=0.3, fp2=0.4, fp3=-0.4, fp4=-0.1, fptimes=seq(from=0, to=550, length.out=5) ) )
oplot.model.data(obsdata, TSMod1class="FuncCombUniformQuasiSinusoid", fname="lcurve_MAPModel.pdf", paramlist=as.list(c(priorform="Gaussian", find.peak.sol(postSamp), list(fptimes=seq(from=0, to=550, length.out=5)))) )
oplot.OUprocess.data(obsdata, fname="lcurve_OUprocess.pdf", yzero=FALSE, plotResid=FALSE, paramlist=list(relax=2.93, diffcon=0.106, offset=0.97, zstartmean=0.94, zstartsd=0.11) )
oplot.OUprocess.data(obsdata, fname="lcurve_OUprocess.pdf", selInd=kfoldCV$partInd[[1]], yzero=FALSE, paramlist=list(relax=19, diffcon=20, offset=0, zstartmean=0, zstartsd=0) )


##### Plot/analyse posterior sampling results (kfoldCV or posterior sampling with full data set)

# Calculate statistics
calc.post.stats(postSamp, thin=1)
calc.post.stats.kfoldcv(kfoldCV)
find.peak.sol(postSamp)
find.peak.sol.kfoldcv(kfoldCV)

# 1D posterior and chain plots
plot.postSamp1d(postSamp, mfrow=c(4,2), plotsize=c(5,7), fname="postSamp1d.pdf")
plot.postSamp1d(postSamp, priorSamp=NULL, Nrej=0, thin=1, Ntrace=1e3, Ndense=2^12, xlim=list(NULL, c(0,0.2), NULL), mfrow=c(3,2), plotsize=c(5,5))
plot.postSamp1d(kfoldCV,  Ntrace=1e3, Ndense=2^12, mfrow=c(5,2), plotsize=c(5,7), fname="kfoldCV.pdf")
plot.postSamp1d(kfoldCV, xlim=list(c(0,30), c(0,0.3), c(0,2), c(0,2), c(0,1)), Ntrace=1e3, Ndense=2^12, mfrow=c(5,2), plotsize=c(5,7), fname="kfoldCV.pdf")

# Plot autocorrelation function of the each chain (parameter) for each partition
plot.chain.acf(kfoldCV, Nrej=100, thin=10, lag.max=10, fname="chain_acf.pdf")

# Recalculate likelihood or KfoldCV likelihood for each partition, possibly using thinning or an hfac
recalc.like(postSamp, hfac=400)
recalc.kfoldcv.like(kfoldCV, thin=10)

# Plot parallel tempering chains
plot.pt.chains(postSampAll, thetaFlag)

# Show thermodynamic integration
plot(beta, cmll, xlab="beta", ylab="mean log (base e) likelihood")
abline(h=logPTEv, col="red")


##### emcee

plot.emcee.walkers(postSampAll, thetaFlag=thetaFlag, Ntrace=500, xlim=NULL, mfcol=c(7,2), plotsize=c(5,7), fname="emcee_walkers.pdf")
plot.emcee.walkers(kfoldCV, part=1, Nwalker=Nwalker, thetaFlag=thetaFlag, Ntrace=500, xlim=NULL, mfcol=c(7,2), plotsize=c(5,7), fname="emcee_walkers.pdf")
plot.walker.acf(postSampAll, thetaFlag, thin=1, lag.max=NULL, ylim=NULL, fname="emcee_walker_acf.pdf")


##### Plot posterior samples with prior samples overplotted

# Draw from priors, and drop the fixed parameters and put the adjustable ones in the same order as postSamp
Npriorsamp <- 1e5
p1 <- sample.prior.theta(TSMod1, theta1Flag, assign.partial(TSMod1@theta1, theta1Fixed), Npriorsamp)[,names(which(theta1Flag))]
p2 <- sample.prior.theta(TSMod2, theta2Flag, assign.partial(TSMod2@theta2, theta2Fixed), Npriorsamp)[,names(which(theta2Flag))]
p3 <- sample.prior.theta(TSMod3, theta3Flag, assign.partial(TSMod3@theta3, theta3Fixed), Npriorsamp)[,names(which(theta3Flag))]
priorSamp <- cbind(p1,p2,p3)
plot.postSamp1d(kfoldCV, priorSamp, Nrej=0, thin=1, Nthinblock=1, Ntrace=200, Ndense=2^12, xlim=list(NULL, NULL, c(0,1), c(-1,+1), c(-1,+1), c(-1,+1), c(-1,+1)), pbound=c(FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE), mfrow=c(7,2), plotsize=c(5,7), fname="kfoldCVpost1Dwithprior.pdf")
#plot.postSamp1d(postSamp, priorSamp, Nrej=0, thin=1, Nthinblock=1, Ntrace=1e3, Ndense=2^12, mfrow=c(4,2), plotsize=c(5,7), fname="postSamp1Dwithprior.pdf")
rm(priorSamp,p1,p2,p3)


##### Calculate likelihood of OU process at specific parameters

paramlist=list(relax=10, diffcon=20, offset=35, zstartmean=0, zstartsd=0) 
#
MeasMod <- new("TimeUniformSignalGaussian")
TSMod2  <- new("ProbOUprocess")
TSMod2@theta2$relax      <- paramlist$relax
TSMod2@theta2$diffcon    <- paramlist$diffcon
TSMod2@theta2$offset     <- paramlist$offset
TSMod2@theta2$zstartmean <- paramlist$zstartmean
TSMod2@theta2$zstartsd   <- paramlist$zstartsd
TSMod3  <- new("ProbUniform")
TSMod3@theta3$tmin <-   0
TSMod3@theta3$tmax <- 100
#  
calc.like.OUprocess(obsdata, 1:nrow(obsdata), TSMod2, TSMod3)
oplot.OUprocess.data(obsdata, fname="offset35zstartmean00zstartsd00.pdf", yzero=FALSE, paramlist=paramlist)


##### Plot FuncQuasiSinusoid - convert into a function maybe

fptimes <- seq(from=0, to=550, length.out=5) # as provided in initialization
fpX <- c(+0.3, 0.4, -0.4, -0.1)
#
TSMod1 <- new("FuncQuasiSinusoid", fptimes=fptimes)
TSMod1@theta1$fp1    <- fpX[1]
TSMod1@theta1$fp2    <- fpX[2]
TSMod1@theta1$fp3    <- fpX[3]
TSMod1@theta1$fp4    <- fpX[4]
TSMod1@theta1$sinamp <- 315
TSMod1@theta1$freq   <- 1/62
TSMod1@theta1$phase  <- 5.2/(2*pi)
tplot <- seq(from=min(fptimes), to=max(fptimes), length.out=200) 
TSMod1@t <- tplot
#
par(mfrow=c(2,1), mar=c(3.5,3.5,1.0,0.5), oma=c(1,1,1,1), mgp=c(2.0,0.6,0))
plot(tplot, eval.mod(TSMod1), type="l", xlab="time", ylab="model prediction") 
vphase <- splint(fptimes, c(0, fpX), xgrid=tplot)
plot(tplot, vphase, type="l", xlab="time", ylab="vphase")
points(fptimes, c(0, fpX), col="red")
abline(h=0, lty=2)
