##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: calc_evidence.R, v20
# Calculate model evidence using prior sampling. Called by run_ctsmod.R

library(doMC)
# Use of doRNG makes no difference, presumably because I am not expecting the state of the RNG
# as it is before the foreach loop to be propagated into the foreach loop. This either implies that the
# integration routines used in calc.like() do not use a RNG, or they seed it differently somehow...

# Return a Nevsamp * (2+p) matrix with named columns, where the columns are
# 1:  log10 prior PDF
# 2:  log10 likelihood
# 3+: the adjustable parameters, i.e. thetaFlag[thetaFlag], in that order
# Column names are set at end of function.
# Note: all parameters passed implicitly, i.e. using R's implicit static scoping rule
# This function writes to the output file pointed to by outEvidence (DISABLED from v16)
calc.evidence <- function() {

  cat("Entering evidence calculation\n")

  ##### Check/set up sampling numbers and parallelization numbers

  if(Nevsamp  <=1) {stop("Nevsamp must be at least 2")}
  if(Nparallel<=1) {stop("Nparallel must be at least 2")}
  if(Nparallel>Nevsamp) {stop("Must have Nparallel<=Nevsamp")}
  Nouter <- Nevsamp %/%  Nparallel
  cat("Using Nouter =", Nouter, "and Nparallel =", Nparallel, "\n")
  if(Nouter*Nparallel != Nevsamp) {
    Nevsamp <- Nouter*Nparallel
    cat("  so Nevsamp has been changed to", Nevsamp, "\n")
  }

  ##### Draw samples from prior (and/or set fixed parameters)

  cat("Sampling the prior\n")
  # Construct a data frame with Nevsamp rows and length(TSModX@thetaX) columns containing all parameter samples.
  # For adjustable parameters, sample from the prior. For fixed parameters, set the fixed values.
  # assign.partial() checks the names in the thetaXFixed list are appropriate to TSModX@thetaX,
  # This check is actually redundant given the earlier checks which ensured that all fixed parameters
  # are already defined in thetaXFixed. The sample.prior.theta functions check that TSModX$alphaX
  # has been assigned values before we try to draw from the prior
  priorSamplesTSMod1 <- sample.prior.theta(TSMod1, theta1Flag, assign.partial(TSMod1@theta1, theta1Fixed), Nevsamp)
  priorSamplesTSMod2 <- sample.prior.theta(TSMod2, theta2Flag, assign.partial(TSMod2@theta2, theta2Fixed), Nevsamp)
  priorSamplesTSMod3 <- sample.prior.theta(TSMod3, theta3Flag, assign.partial(TSMod3@theta3, theta3Fixed), Nevsamp)
  
  ##### Calculate evidence

  # k=1:Nouter             the outer, non-parallel loop
  # p=1:Nparallel          the inner, parallel foreach loop
  # n=(k-1)*Nparallel + p  counts over the prior samples
  # j=1:nrow(obsdata)      events counter

  if(Ncores>0) {registerDoMC(Ncores)} else {registerDoMC()} # If specified, allocate Ncores, otherwise use as many as available
  thetaNames <- noquote(names(c(TSMod1@theta1, TSMod2@theta2, TSMod3@theta3))) # order is as in class prototypes
  
  cat("Calculating the evidence using", getDoParWorkers(), "(possibly virtual) cores.\nOutput columns:\n")
  if(is.finite(verboseE)) {cat("iteration # of #   logPrior    logLike :", thetaNames, ": time elapsed, expected duration\n")}
  startTime <- Sys.time()
  evCalc <- matrix(data=NA, nrow=Nevsamp, ncol=2+length(which(thetaFlag)))
  colnames(evCalc) <- c("logPrior", "logLike", names(which(thetaFlag)))
  for(k in 1:Nouter) {
    
    loopy <- foreach(p = 1:Nparallel, .combine="rbind") %dopar% { # %do% will do loop in series
      n <- (k-1)*Nparallel + p
      # Put prior samples into objects. Note that set.theta() uses the column names of priorSamplesTSModX
      # rather than TSModX@thetaX, thus ensuring that the data are assigned to the correct parameters
      # (i.e. column order returned above by sample.prior.theta is irrelevant)
      TSMod1@theta1 <- set.theta(priorSamplesTSMod1, n)
      TSMod2@theta2 <- set.theta(priorSamplesTSMod2, n)
      TSMod3@theta3 <- set.theta(priorSamplesTSMod3, n)
      thetaAdj <- as.numeric( c(TSMod1@theta1, TSMod2@theta2, TSMod3@theta3)[names(which(thetaFlag))] )
      # v15: names(which(thetaFlag)) selects the adjustable theta, in the order of thetaFlag
      logPrior <- log10( eval.prior.theta(TSMod1) * eval.prior.theta(TSMod2) * eval.prior.theta(TSMod3) )
      logLike  <- calc.like(obsdata, ind=1:nrow(obsdata), TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n) # log base 10
      if( is.finite(verboseE) && (n %% verboseE == 0) ) {
        dump.evidence.diagnostics(n, Nevsamp, logPrior, logLike, TSMod1, TSMod2, TSMod3, startTime)
      }
      c(logPrior, logLike, thetaAdj)
    }
    evCalc[((k-1)*Nparallel + 1):(k*Nparallel),] <- loopy
    
  }
  
  registerDoSEQ() # return to serial processing
  if(any(is.na(evCalc))) {warning("NA/NaN present in evCalc")}
  colnames(evCalc) <- c("logPrior", "logLike", names(which(thetaFlag)))
  rownames(evCalc) <- c(1:Nevsamp)
  return(evCalc)
  
}



####################
#################### Screen/file printing functions
####################

# Print various diagnostics to screen during likelihood calculations
dump.evidence.diagnostics <- function(n, Nevsamp, logPrior, logLike, TSMod1, TSMod2, TSMod3, startTime) {
  sdump1 <- noquote( formatC(n,        format="d", digits=5, flag="") )
  sdump2 <- noquote( formatC(Nevsamp,  format="g", digits=5, flag="") )
  sdump3 <- noquote( formatC(logPrior, format="e", width=10, digits=3, flag="") )
  sdump4 <- noquote( formatC(logLike,  format="e", width=10, digits=3, flag="") )
  # order of parameters printed within each thetaX in sdump5 is determined by order of writing by sample.prior.theta()
  sdump5 <- noquote( formatC(as.numeric(c(TSMod1@theta1, TSMod2@theta2, TSMod3@theta3)), format="g", width=8, digits=3, flag="") )
  est <- est.time(startTime, n, Nevsamp)
  sdump6 <- noquote( formatC(est$elapsedTime,  format="f", width=5, digits=2, flag="") )
  sdump7 <- noquote( formatC(est$expectedDuration,  format="f", width=5, digits=2, flag="") )
  cat(sdump1, "of", sdump2, sdump3, sdump4, ":", sdump5, ":", sdump6, est$elapsedTimeUnits, sdump7, est$expectedDurationUnits, "\n")
  return(TRUE) # ????
}

