##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: ctsmod_likefunctions.R, v21
# Likelihood evaluation functions for ctsmod. 

# R2Cuba.mod is the R2Cuba package modified to remove an annoying cat("\n") when verbose==0
# See ~/doc/R_install for how to achieve this. Source code is in R2Cuba.mod/
#if(system("hostname", intern=TRUE)=='ganges.local') {
#  library(R2Cuba.mod) 
#} else {
#  library(R2Cuba) 
#}
library(R2Cuba)


#################### 
#################### Likelihood calculations
#################### 


# Given 2D data matrix obsdata, calculate the likelihood for the events in the row numbers
# specified by the vector ind, given the model parameters specified in TSModX@thetaX and for
# measurement model MeasMod. tzlim and integopt are used for some of the integration methods
# in the event likelihood calculation.
# n is some row number from which the parameters were set (i.e. from a matrix of prior
# or posterior samples). thetaFlag and n are only used in a warning report.
# Return the log10 likelihood, i.e. the probability of obsdata for given TSMod parameters.
calc.like <- function(obsdata, ind, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n) {
  if(!all(is.element(ind, 1:nrow(obsdata)))) {stop("ind contains indices not in 1:nrow(obsdata)")}
  if(class(TSMod2)=="ProbOUprocess" || class(TSMod2)=="ProbWienerprocess") {
    logLike <- calc.like.FullyStochasticprocess(obsdata, ind, TSMod2, TSMod3)
  } else {
    logLike <- calc.like.isolated(matrix(obsdata[ind,],ncol=4,byrow=FALSE), TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n)
  }
  return(logLike)
}


# Calculate the likelihood for the (normal) case that the calculations for each event are isolated
# from one another.
# obsdata must be a 2D numerical matrix, even if it only has one row.
# n is some row number from which the parameters were set (i.e. from a matrix of prior
# or posterior samples). It may be NA. thetaFlag and n are only used in a warning report.
# Return the log10 likelihood, i.e. the probability of obsdata for given TSMod parameters.
calc.like.isolated <- function(obsdata, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n) {  
  eventLike <- vector(mode="numeric", length=nrow(obsdata))
  for(j in 1:length(eventLike)) {
    MeasMod@obsdata <- obsdata[j,]
    integout <- calc.event.like(TSMod1, TSMod2, TSMod3, MeasMod, tzlim, integopt)
    eventLike[j] <- integout$value
    if(!is.null(integout$errorMessage)) {
      thetaAdj <- as.numeric(c(TSMod1@theta1, TSMod2@theta2, TSMod3@theta3)[thetaFlag])
      warning("calc.like.isolated(): Sample=", n, " event=", j, " thetaAdj=", list(thetaAdj), " eventLike=", eventLike[j], " ", integout$errorMessage)
    }
  }
  logLike  <- sum(log10(eventLike))
  return(logLike)
}


# Calculate single event likelihood, i.e. 2D integral over t_j, z_j of TSMod2*TSMod3*MeasMod
# This is generic for any TSmod (achieved using the OO feature of operator overloading)
#
# Calculates the single event likelihood, i.e. the 2D integration over t_j, z_j, using a numerical algorithm.
# In special cases this can be simplified to a 1D integration or a 0D integration (see description TN).
# Returns a two component list: the value (real), and a possible error message (character).
# If no error, the latter is NULL.
# If the integration fails, an error message is returned, and the likelihood set to zero.
# If the integral is negative, the likelihood is set to zero. If it is more negative than the magnitude of the numerical error,
# then an error message is returned.
# tzlim = c(t.lower, t.upper, z.lower, z.upper)
# Internally:
# Rather than using a long switch statement, all the different ways of integration are inside "if" clauses,
# which set integdone=TRUE at end, which is then used to skip the other ways of doing integration
# All different ways of integration must set integout = list(value, ifail, abs.error)
# where ifail=1/0 for success/failure. Set abs.error=0 if there is no estimate of the uncertainty in the result.
# The list integout may include other tags, which are ignored.
calc.event.like <- function(TSMod1, TSMod2, TSMod3, MeasMod, tzlim, integopt) {
  integdone <- FALSE
  errorMessage <- NULL

  # If time uncertainties are negligible compared to TSMod variation time scale (as set by second "if")
  # then for certain (common) models (specified in first "if") we can use 0D integration
  if( (class(TSMod2)=="ProbGaussian" || class(TSMod2)=="BypassTSMod2") && (class(MeasMod)=="TwoGaussian" || class(MeasMod)=="TimeUniformSignalGaussian") ) {
    if( (MeasMod@obsdata[2]==0 || integopt$negtsd) ) {
      integout <- list(value=calc.event.like.negtsd(TSMod1, TSMod2, TSMod3, MeasMod), ifail=0, abs.error=0)
      integdone <- TRUE
    }
  }

  # If s.sd=y.sd=0 and we have a stochastic signal component (i.e. not doing TSMod2 bypass), then
  # we can bypass the measurement model and do a 0D integration
  if(!integdone && class(TSMod2)!="BypassTSMod2" && MeasMod@obsdata[2]==0 && MeasMod@obsdata[4]==0) {
    integout <- list(value=calc.event.like.zerobothsd(MeasMod@obsdata[c(1,3)], TSMod1, TSMod2, TSMod3), ifail=0, abs.error=0)
    integdone <- TRUE
  }
  
  # Doing TSMod2 bypass means we just have to do a 1D integration over t; here with integrate{stats}
  # integrate returns a list including tags "value" and "abs.error"
  if(!integdone && class(TSMod2)=="BypassTSMod2") { 
    # f in integrate() takes a vector as its first argument and return a corresponding vector
    if(integopt$shrink=='both' || integopt$shrink=='only_t') {
      tzlim <- shrink.tlim(MeasMod, tzlim)
    }
    integout <- integrate(f=calc.event.like.integrand.2, lower=tzlim[1], upper=tzlim[2], subdivisions=1e3, rel.tol=1e-8, abs.tol=1e-8, stop.on.error=FALSE, TSMod1, TSMod3, MeasMod)
    # add required tag ifail
    if(integout$message=="OK") {
      integout$ifail <- 0
    } else {
      integout$ifail <- 1
    }
    integdone <- TRUE
  } 

  # Full 2D integration with cuhre{R2Cuba}
  # integrand in cuhre takes a vector as its first argument of length ndim
  # cuhre() returns a list including tags "value", "ifail" and "abs.error"
  if(!integdone) {
    if(integopt$shrink=='only_t' || integopt$shrink=='both') {
      tzlim <- shrink.tlim(MeasMod, tzlim)
    }
    if(integopt$shrink=='only_z' || integopt$shrink=='both') {
      tzlim <- shrink.zlim(MeasMod, tzlim)
    }
    if(class(TSMod2)=="ProbOUprocNaive" || class(TSMod2)=="ProbIntOUprocNaive") { # TSMod1 is bypassed
      integout <- cuhre(integrand=calc.event.like.integrand.3, ncomp=1, ndim=2, lower=tzlim[c(1,3)], upper=tzlim[c(2,4)], rel.tol=1e-4, abs.tol=0, flags=list(verbose=0, final=1), min.eval=0, max.eval=1e6, key=0, TSMod2, TSMod3, MeasMod)
    } else {
      integout <- cuhre(integrand=calc.event.like.integrand, ncomp=1, ndim=2, lower=tzlim[c(1,3)], upper=tzlim[c(2,4)], rel.tol=1e-4, abs.tol=0, flags=list(verbose=0, final=1), min.eval=0, max.eval=1e6, key=0, TSMod1, TSMod2, TSMod3, MeasMod)
    }
    integdone <- TRUE
  }
  
  # Deal with errors
  # cat("integout(value, ifail, abs.error): ", integout$value, integout$ifail, integout$abs.error, "\n")
  if(!integdone) {stop("No integration/calculation selection criteria met: coding or set-up error")}
  if(integout$ifail!=0) {
    errorMessage <- c("Integration failed. Event likelihood set to zero. Error message: ", integout$message, ". Function call: ", integout$call, sep="")
    integout$value <- 0
  }
  if(integout$value < 0) {
    if(abs(integout$value)>abs(integout$abs.error)) { 
      errorMessage <- "Event likelihood more negative than numerical uncertainty, but set to zero."
    } else {
      errorMessage <- "Event likelihood negative but smaller than numerical uncertainty. Have set to zero."
    }
    integout$value <- 0
  }

  return(list(value=integout$value, errorMessage=errorMessage))  
}


# Function returning the event likelihood integrand for use in calc.event.like()
# for the "general" case of TS model being specified in all three parts, TSModX X=1,2,3.
# voi (variables of integration) is a 2-element vector, (t,z)
# All other parameters of all objects passed must be set in advance
# Remember that "<-" is assignment of value,
# so we have to "refill" TSMod2@TSMod1SC=TSMod1 after setting TSMod1@t
calc.event.like.integrand <- function(voi, TSMod1, TSMod2, TSMod3, MeasMod) {
  TSMod1@t  <- voi[1]
  TSMod2@z  <- voi[2]
  TSMod2@TSMod1SC <- TSMod1
  TSMod3@t  <- voi[1]
  MeasMod@t <- voi[1]
  MeasMod@z <- voi[2]
  #cat("Evaluations: TSMod2:", eval.mod(TSMod2), "TSMod3:", eval.mod(TSMod3), "MeasMod:", eval.mod(MeasMod), "\n")
  # cat("Evaluations: t =", MeasMod@t, "s =", MeasMod@obsdata[1], "z =", MeasMod@z, "y =", MeasMod@obsdata[3], "MeasMod:", eval.mod(MeasMod), "\n")
  return( eval.mod(TSMod2)*eval.mod(TSMod3)*eval.mod(MeasMod) )
}



# Function returning the event likelihood integrand for use in calc.event.like()
# for the case of purely stochastic models (TSMod1 bypassed).
# voi (variables of integration) is a 2-element vector, (t,z)
# All other parameters of all objects passed must be set in advance
calc.event.like.integrand.3 <- function(voi, TSMod2, TSMod3, MeasMod) {
  TSMod2@t  <- voi[1]
  TSMod2@z  <- voi[2]
  TSMod3@t  <- voi[1]
  MeasMod@t <- voi[1]
  MeasMod@z <- voi[2]
  return( eval.mod(TSMod2)*eval.mod(TSMod3)*eval.mod(MeasMod) )
}


# Function returning the event likelihood integrand for use in calc.event.like()
# for the specific case of TSMod2 not being used (the TSMod2 bypass),
# in which case this is a function only of t (and not z).
# if t (the variable of integration) is a scalar, then a scalar is returned.
# if it is a vector, then a vector of the same length is returned.
# Note that in that case eval.mod(TSMod1) returns a vector of this length.
calc.event.like.integrand.2 <- function(t, TSMod1, TSMod3, MeasMod) {
  TSMod1@t  <- t
  TSMod3@t  <- t
  MeasMod@t <- t
  MeasMod@z <- eval.mod(TSMod1)
  return( eval.mod(TSMod3)*eval.mod(MeasMod) )
}


# Calculate event likelihood in the case that time uncertainties are negligible compared to TSMod variation time scale,
# in which case we can do the  "0D" integration. This is only appropriate for certain (but common) models,
# namely (TSMod2=ProbGaussian or BypassTSMod2) together with (MeasMod=TwoGaussian or TimeUniformSignalGaussian)
calc.event.like.negtsd <- function(TSMod1, TSMod2, TSMod3, MeasMod) {
  TSMod1@t <- MeasMod@obsdata[1]
  TSMod3@t <- MeasMod@obsdata[1]
  if(class(TSMod2)=="ProbGaussian") {
    sdTot <- sqrt(MeasMod@obsdata[4]^2+TSMod2@theta2$sd^2)
  } else { # class(TSMod2)=="BypassTSMod2" is assumed
    sdTot <- MeasMod@obsdata[4]
  }
  val <- eval.mod(TSMod3) * dnorm(x=MeasMod@obsdata[3], mean=eval.mod(TSMod1), sd=sdTot)
  return(val)
}


# Calculate event likelihood under the assumption that both the signal and time uncertainties
# are negligibly small, and we have some stochastic signal component (i.e. not doing TSMod2 bypass).
# Likelihood is then a 0D integration; function evaluated at 2D position tzval.
calc.event.like.zerobothsd <- function(tzval, TSMod1, TSMod2, TSMod3) {
  if(class(TSMod2)=="ProbOUprocNaive" || class(TSMod2)=="ProbIntOUprocNaive") { # actually for any fully stochastic model (TSMod1 bypassed)
    TSMod2@t  <- tzval[1]
  } else {
    TSMod1@t  <- tzval[1]
    TSMod2@TSMod1SC <- TSMod1
  }
  TSMod2@z  <- tzval[2]
  TSMod3@t  <- tzval[1]
  return( eval.mod(TSMod2)*eval.mod(TSMod3) )
}

# Calculate the likelihood for a "fully stochastic process", given the model parameters in TSModX@thetaX.
# This is defined as a process for which the likelihood can be calculated using the procedure
# outlined in the appendix of the ctsmod paper (Bailer-Jones 2012). This currently applies to models
# ProbOUprocess, ProbWienerprocess.
# In these processes we have to propagate estimates of the process parameters along the chain of events,
# and for this we use all of the events in obsdata. However, the likelihood is calculated
# and returned only for the events specified in ind.
# Note that these processes have made particular assumptions about MeasMod (which is why it is not passed).
# This function uses vectorized calculations.
# Return the log10 likelihood, i.e. the probability of obsdata for given TSMod parameters.
calc.like.FullyStochasticprocess <- function(obsdata, ind, TSMod2, TSMod3) {
  conv  <- 1/log(10)
  procParam <- eval.mod(TSMod2, obsdata) # gives 2-column data frame of zMean, zVar at all epochs
  TSMod3@t <- obsdata[,1]
  yMean <- procParam$zMean
  yVar  <- obsdata[,4]^2 + procParam$zVar
  logEventLike <- log10(eval.mod(TSMod3)[ind]) + conv*dnorm(x=obsdata[ind,3], mean=yMean[ind], sd=sqrt(yVar[ind]), log=TRUE)
  # print(cbind( eval.mod(TSMod3)[ind], obsdata[ind,3], yMean[ind], sqrt(yVar[ind]), dnorm(x=obsdata[ind,3], mean=yMean[ind], sd=sqrt(yVar[ind]), log=TRUE), logEventLike ))
  return( sum(logEventLike) )
}



#################### 
#################### Limit shrinking
#################### 

# For most MeasMod, we know in advance at what values eval.mod will return
# a negligible contribution. This is exploited here to return narrower limits on t and/or z
# for the event likelihood integration, in order to save time/improve accuracy.
# Note: limits will never be expanded!
# Returns tzlim = c(t.lower, t.upper, z.lower, z.upper)
# Currently implemented for: TwoGaussian, TimeUniformSignalGaussian

# Shrink t limits. Returns tzlim, possibly changing positions 1 and 2 (not 3 or 4)
shrink.tlim <- function(MeasMod, tzlim) {
  if(class(MeasMod)=='TwoGaussian') {
    sigLim <- 5.5 # P=1e-7
    tmin <- MeasMod@obsdata[1] - sigLim*MeasMod@obsdata[2]
    if(tzlim[1] < tmin) {tzlim[1] <- tmin}
    tmax <- MeasMod@obsdata[1] + sigLim*MeasMod@obsdata[2]
    if(tzlim[2] > tmax) {tzlim[2] <- tmax}
  }
  if(class(MeasMod)=='TimeUniformSignalGaussian') {
    tmin <- MeasMod@obsdata[1] - (sqrt(12)/2)*MeasMod@obsdata[2]
    if(tzlim[1] < tmin) {tzlim[1] <- tmin}
    tmax <- MeasMod@obsdata[1] + (sqrt(12)/2)*MeasMod@obsdata[2]
    if(tzlim[2] > tmax) {tzlim[2] <- tmax}
  }
  return(tzlim)
}

# Shrink z limits. Returns tzlim,  possibly changing positions 3 and 4 (not 1 or 2)
shrink.zlim <- function(MeasMod, tzlim) {
  if(class(MeasMod)=='TwoGaussian') {
    sigLim <- 5.5 # P=1e-7
    zmin <- MeasMod@obsdata[3] - sigLim*MeasMod@obsdata[4]
    if(tzlim[3] < zmin) {tzlim[3] <- zmin}
    zmax <- MeasMod@obsdata[3] + sigLim*MeasMod@obsdata[4]
    if(tzlim[4] > zmax) {tzlim[4] <- zmax}
  }
  if(class(MeasMod)=='TimeUniformSignalGaussian') {
    zmin <- MeasMod@obsdata[3] - (sqrt(12)/2)*MeasMod@obsdata[4]
    if(tzlim[3] < zmin) {tzlim[3] <- zmin}
    zmax <- MeasMod@obsdata[3] + (sqrt(12)/2)*MeasMod@obsdata[4]
    if(tzlim[4] > zmax) {tzlim[4] <- zmax}
  }
  return(tzlim)
}


