##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: ctsmod_models.R, v23
# Model definitions (classes and methods) for ctsmod. See notes in run_ctsmod.R in particular.

library(fields) # required by FuncQuasiSinusoid

####################
#################### Generic functions (to be extended by methods of classes)
####################

# Evaluate TSMod or MeasMod.
# May return a scalar, vector or a dataframe (see specific implementations).
# Return value is typically a function value (for TSMod1) or a probability
# (for TSMod2, TSMod3, MeasMod) or even parameters of a PDF (for TSMod2).
eval.mod <- function(object, ...) {}
setGeneric("eval.mod")

# Evaluate prior PDF P(theta ; alpha). In general not normalized?
# Prior for a parameter is set to 1 is any of its alpha hyperparameters are null.
# Returns a real scalar
eval.prior.theta <- function(object) {}
setGeneric("eval.prior.theta")

# Drawn n samples from prior PDF P(theta; alpha)
# It is an error (stop) if any of the alpha hyperparameters are null for an adjustable theta parameter,
# Returns a dataframe of dimension n * length(theta) with named columns (theta parameters)
# For those theta with thetaFlag=FALSE, just repeat the corresponding value of thetaFixed
# in that column. thetaFlag is a named logical vector, thetaFixed is a named real vector.
# objects is of class TSMod
sample.prior.theta <- function(object, thetaFlag, thetaFixed, n) {}
setGeneric("sample.prior.theta")

# Note: the logical consistency of many of the parameters used in the methods
# is assumed to have been checked before using the functions.
# However, the values of the alpha and (fixed) theta parameters are not checked anywhere:
# this is model specific so would need a lot of code. Some inappropriate values may well
# gives errors, although these might be hard to track. Be careful!

# Map true theta to tranformed(theta), which has range +/- infinity,
# for those theta parameter names which are TRUE in thetaFlag.
# Otherwise leave unchanged. 
# This is used by the MCMC function to ensure that all adjustable parameters have an
# infinite range, so that a multivariate normal is an appropriate proposal distribution.
# The standard transformation types depend on the prior:
# theta ~ Gaussian =>  -Inf:Inf  =>  no transformation required
# theta ~ Gamma    =>     0:Inf  =>  log10(theta)   inverse is 10^()
# theta ~ Uni(0,1) =>     0:1    =>  logit          inverse is sigmoid()
# Returns object@theta1 for this specific class.Note that the return is like this:
#   object@theta1[names(new("FuncSinusoid")@theta1)]
# and not this
#    return(object@theta1)
# because object@theta1 may contain more parameters than belong to this class,
# i.e. if object is actually of an inherited class, e.g. FuncCombUniformSinusoid.
trans.theta <- function(object, thetaFlag) {}
setGeneric("trans.theta")

# Map transformed(theta, thetaFlag) back to true theta,
# for those theta parameter names which are TRUE in thetaFlag.
# Otherwise leave unchanged.
# Returns a list object@theta for this class - see notes for trans.theta() above.
invtrans.theta <- function(object, thetaFlag) {}
setGeneric("invtrans.theta")

# Compute the Jacobian (a scalar) corresponding to the transformation in trans.theta().
# As all transformations are 1D, the Jacobian is just the product of individual
# transformations. Only calculate for those variables where thetaFlag is TRUE,
# otherwise the Jacobian is unity. The variables passed in through object should
# be the transformed ones, i.e. the result of trans.theta().
jacobian.theta <- function(object, thetaFlag) {
  return(1)
}
setGeneric("jacobian.theta") # note that the above is the default method


####################
#################### Virtual classes
####################

# Bypass
# This is a dummy model which just indicates that we want to bypass the model. It has one parameter, "dummy".
# The model will never be evaluated, so has no eval.mod method. An evaluation of the prior returns unity
# and samples drawn from the prior are NA (which is used by the code to indicate that we are bypassing this model).
# This is implemented as a virtual class. Objects cannot be instantiated because it defines
# no slots (no representation). It exits purely to provide standard methods to be inherited by
# TSModX-specific bypass classes which define the necessary slots. These must have:
#   representation=representation(thetaX="list", alphaX="list")
#   prototype=prototype(thetaX=list(dummy=NULL) 
setClass("BypassSC")
#
eval.prior.theta.BypassSC <- function(object) {
  return(1)
}
setMethod(eval.prior.theta, signature=c("BypassSC"), definition=eval.prior.theta.BypassSC)
#
sample.prior.theta.BypassSC <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(dummy=NULL)
  samples$dummy <- rep.int(times=n, x=NA)
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("BypassSC"), definition=sample.prior.theta.BypassSC)
#
trans.theta.BypassSC <- function(object, thetaFlag) {
  return(list(dummy=NULL))
}
setMethod(trans.theta, signature=c("BypassSC"), definition=trans.theta.BypassSC)
#
invtrans.theta.BypassSC <- function(object, thetaFlag) {
  return(list(dummy=NULL))
}
setMethod(invtrans.theta, signature=c("BypassSC"), definition=invtrans.theta.BypassSC)



####################
#################### TSMod1. TS model, deterministic part (theta1). Returns f(t, theta1 ; TSMod1) = <z(t)>
####################

# For all implementations:
# eval.mod             Return scalar, or vector of length(object@t)
# eval.prior.theta     Return scalar
# sample.priortheta    Return vector of length n (one of the specified arguments)

# superclass (This superclass is required by the TSMod2 definitions)
setClass("TSMod1SC", representation=representation(t="numeric", theta1="list", alpha1="list"))


#################### FuncUniform

# FuncUniform (offset)
# Choice of prior selected by specifying priorform, e.g. in new("FuncUniform", priorform="Gamma")
# P(offset) ~ Gamma(shape, scale)  when priorform="Gamma" 
# P(offset) ~ Gaussian(mean, sd)   when priorform="Gaussian" [default]
setClass("FuncUniform", contains="TSMod1SC", representation=representation(priorform="character"), prototype=prototype(t=NULL, theta1=list(offset=NULL), priorform=NULL, alpha1=list(offset=NULL) ) )
#
init.FuncUniform <- function(.Object, priorform="Gaussian") {
  if(priorform=="Gamma") {
    .Object@priorform <- "Gamma"
    .Object@alpha1$offset <- list(shape=NULL, scale=NULL)
  } else {
    .Object@priorform <- "Gaussian"
    .Object@alpha1$offset <- list(mean=NULL, sd=NULL)
  }
  .Object
}
setMethod("initialize", signature="FuncUniform", definition=init.FuncUniform)
#
eval.mod.FuncUniform <- function(object) {
  return( rep(object@theta1$offset, length(object@t)) )
}
setMethod(eval.mod, signature=c("FuncUniform"), definition=eval.mod.FuncUniform)
#
eval.prior.theta.FuncUniform <- function(object) {
  if(object@priorform=="Gamma") {
    if(is.null(object@alpha1$offset$shape) || is.null(object@alpha1$offset$scale)) {
      return(1)
    } else {
      return( dgamma(object@theta1$offset, shape=object@alpha1$offset$shape, scale=object@alpha1$offset$scale) )
    }
  } else {
    if(is.null(object@alpha1$offset$mean) || is.null(object@alpha1$offset$sd)) {
      return(1)
    } else {
      return( dnorm(object@theta1$offset, mean=object@alpha1$offset$mean, sd=object@alpha1$offset$sd) )
    }
  }
}
setMethod(eval.prior.theta, signature=c("FuncUniform"), definition=eval.prior.theta.FuncUniform)
#
sample.prior.theta.FuncUniform <- function(object, thetaFlag, thetaFixed, n) {
  samples <- list(offset=NULL)
  if(thetaFlag["offset"]) {
    if(object@priorform=="Gamma") {
      if(is.null(object@alpha1$offset$shape) || is.null(object@alpha1$offset$scale)) {
        stop("Prior parameters for TSMod1@alpha1$offset have not been set")
      }
      samples$offset <- rgamma(n, shape=object@alpha1$offset$shape, scale=object@alpha1$offset$scale)
    } else {
      if(is.null(object@alpha1$offset$mean) || is.null(object@alpha1$offset$sd)) {
        stop("Prior parameters for TSMod1@alpha1$offset have not been set")
      }
      samples$offset <- rnorm(n, mean=object@alpha1$offset$mean, sd=object@alpha1$offset$sd)
    }
  } else {
    samples$offset <- rep.int(times=n, x=thetaFixed$offset)
  }
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("FuncUniform"), definition=sample.prior.theta.FuncUniform)
#
trans.theta.FuncUniform <- function(object, thetaFlag) {
  if(thetaFlag["offset"]) {
    if(object@priorform=="Gamma") {
      object@theta1$offset <- log10(object@theta1$offset)
    }
  }
  return( object@theta1[names(new("FuncUniform")@theta1)] )
}
setMethod(trans.theta, signature=c("FuncUniform"), definition=trans.theta.FuncUniform)
#
invtrans.theta.FuncUniform <- function(object, thetaFlag) {
  if(thetaFlag["offset"]) {
    if(object@priorform=="Gamma") {
      object@theta1$offset <- 10^object@theta1$offset
    }
  }
  return( object@theta1[names(new("FuncUniform")@theta1)] )
}
setMethod(invtrans.theta, signature=c("FuncUniform"), definition=invtrans.theta.FuncUniform)
#
jacobian.theta.FuncUniform <- function(object, thetaFlag) {
  if(thetaFlag["offset"]) {
    if(object@priorform=="Gamma") {
      return( log(10) * exp(object@theta1$offset) )
    }
  } else {
    return(1)  
  }
}
setMethod(jacobian.theta, signature=c("FuncUniform"), definition=jacobian.theta.FuncUniform)


#################### FuncSigmoid

# FuncSigmoid (sigamp, lambda, tzero)
# P(sigamp) ~ Gamma(shape, scale)   P(lambda) ~ N(mean, sd)   P(tzero) ~ N(mean, sd)
setClass("FuncSigmoid", contains="TSMod1SC", prototype=prototype(t=NULL, theta1=list(sigamp=NULL, lambda=NULL, tzero=NULL), alpha1=list(sigamp=list(shape=NULL, scale=NULL), lambda=list(mean=NULL, sd=NULL), tzero=list(mean=NULL, sd=NULL)) ) )
#
eval.mod.FuncSigmoid <- function(object) {
  return( object@theta1$sigamp / ( 1+exp(-(object@t-object@theta1$tzero)/object@theta1$lambda) ) ) # (transforms +Inf and -Inf okay)
}
setMethod(eval.mod, signature=c("FuncSigmoid"), definition=eval.mod.FuncSigmoid)
#
eval.prior.theta.FuncSigmoid <- function(object) {
  sigampPrior <- ifelse( (is.null(object@alpha1$sigamp$shape) || is.null(object@alpha1$sigamp$scale)), 1, dgamma(object@theta1$sigamp, shape=object@alpha1$sigamp$shape, scale=object@alpha1$sigamp$scale) )
  lambdaPrior <- ifelse( (is.null(object@alpha1$lambda$mean) || is.null(object@alpha1$lambda$sd)), 1, dnorm(object@theta1$lambda, mean=object@alpha1$lambda$mean, sd=object@alpha1$lambda$sd) )
  tzeroPrior <- ifelse( (is.null(object@alpha1$tzero$mean) || is.null(object@alpha1$tzero$sd)), 1, dnorm(object@theta1$tzero, mean=object@alpha1$tzero$mean, sd=object@alpha1$tzero$sd) )
  return( prod(sigampPrior, lambdaPrior, tzeroPrior) )
}
setMethod(eval.prior.theta, signature=c("FuncSigmoid"), definition=eval.prior.theta.FuncSigmoid)
#
sample.prior.theta.FuncSigmoid <- function(object, thetaFlag, thetaFixed, n) {
  samples <- list(sigamp=NULL, lambda=NULL, tzero=NULL)
  if(thetaFlag["sigamp"]) {
    if(is.null(object@alpha1$sigamp$shape) || is.null(object@alpha1$sigamp$scale)) {
      stop("Prior parameters for TSMod1@alpha1$sigamp have not been set")
    }
    samples$sigamp <- rgamma(n, shape=object@alpha1$sigamp$shape, scale=object@alpha1$sigamp$scale)
  } else {
    samples$sigamp <- rep.int(times=n, x=thetaFixed$sigamp)
  }
  if(thetaFlag["lambda"]) {
    if(is.null(object@alpha1$lambda$mean) || is.null(object@alpha1$lambda$sd)) {
      stop("Prior parameters for TSMod1@alpha1$lambda have not been set")
    }
    samples$lambda <- rnorm(n, mean=object@alpha1$lambda$mean, sd=object@alpha1$lambda$sd)
  } else {
    samples$lambda <- rep.int(times=n, x=thetaFixed$lambda)
  } 
  if(thetaFlag["tzero"]) {
    if(is.null(object@alpha1$tzero$mean) || is.null(object@alpha1$tzero$sd)) {
      stop("Prior parameters for TSMod1@alpha1$tzero have not been set")
    }
    samples$tzero <- rnorm(n, mean=object@alpha1$tzero$mean, sd=object@alpha1$tzero$sd) 
  } else {
    samples$tzero <- rep.int(times=n, x=thetaFixed$tzero)
  }
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("FuncSigmoid"), definition=sample.prior.theta.FuncSigmoid)
#
trans.theta.FuncSigmoid <- function(object, thetaFlag) {
  if(thetaFlag["sigamp"]) {
    object@theta1$sigamp <- log10(object@theta1$sigamp)
  }
  return( object@theta1[names(new("FuncSigmoid")@theta1)] )
}
setMethod(trans.theta, signature=c("FuncSigmoid"), definition=trans.theta.FuncSigmoid)
#
invtrans.theta.FuncSigmoid <- function(object, thetaFlag) {
  if(thetaFlag["sigamp"]) {
    object@theta1$sigamp <- 10^object@theta1$sigamp
  }
  return( object@theta1[names(new("FuncSigmoid")@theta1)] )
}
setMethod(invtrans.theta, signature=c("FuncSigmoid"), definition=invtrans.theta.FuncSigmoid)
#
jacobian.theta.FuncSigmoid <- function(object, thetaFlag) {
  if(thetaFlag["sigamp"]) {
    return( log(10) * exp(object@theta1$sigamp) )
  } else {
    return(1)  
  }
}
setMethod(jacobian.theta, signature=c("FuncSigmoid"), definition=jacobian.theta.FuncSigmoid)


#################### FuncSigmoidZero

# FuncSigmoidZero (sigamp, lambda, tzero)
# P(sigamp) ~ Gamma(shape, scale)   P(lambda) ~ N(mean, sd)   P(tzero) ~ N(mean, sd)
# As FuncSigmoid but symmetric in signal about zero (rather than having zero as the minimum)
setClass("FuncSigmoidZero", contains="FuncSigmoid")
#
eval.mod.FuncSigmoidZero <- function(object) {
  return( ( object@theta1$sigamp / ( 1+exp(-(object@t-object@theta1$tzero)/object@theta1$lambda) ) ) - object@theta1$sigamp/2 )
}
setMethod(eval.mod, signature=c("FuncSigmoidZero"), definition=eval.mod.FuncSigmoidZero)
# No need to explicitly define prior evaluation/sampling or transformation/jacobian methods, as they are identical and inherited



#################### FuncSinusoid

# FuncSinusoid (sinamp, freq, phase)
# P(sinamp) ~ Gamma(shape, scale)  P(freq) ~ Gamma(shape, scale)  P(phase) ~ Uniform()
setClass("FuncSinusoid", contains="TSMod1SC", prototype=prototype(t=NULL, theta1=list(sinamp=NULL, freq=NULL, phase=NULL), alpha1=list(sinamp=list(shape=NULL, scale=NULL), freq=list(shape=NULL, scale=NULL), phase=NULL) ) )
#
eval.mod.FuncSinusoid <- function(object) {
  return( (object@theta1$sinamp/2) * ( cos(2*pi*(object@theta1$freq*object@t + object@theta1$phase)) + 1 ) )
}
setMethod(eval.mod, signature=c("FuncSinusoid"), definition=eval.mod.FuncSinusoid)
#
eval.prior.theta.FuncSinusoid <- function(object) {
  sinampPrior <- ifelse( (is.null(object@alpha1$sinamp$shape) || is.null(object@alpha1$sinamp$scale)), 1, dgamma(object@theta1$sinamp, shape=object@alpha1$sinamp$shape, scale=object@alpha1$sinamp$scale) )
  freqPrior <- ifelse ( (is.null(object@alpha1$freq$shape) || is.null(object@alpha1$freq$scale)), 1, dgamma(object@theta1$freq, shape=object@alpha1$freq$shape, scale=object@alpha1$freq$scale) )
# v20: enforce 0-1 prior on phase
  phasePrior <- dunif(object@theta1$phase, min=0, max=1)
#  phasePrior <- 1
  return( prod(sinampPrior, freqPrior, phasePrior) )
}
setMethod(eval.prior.theta, signature=c("FuncSinusoid"), definition=eval.prior.theta.FuncSinusoid)
#
sample.prior.theta.FuncSinusoid <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(sinamp=NULL, freq=NULL, phase=NULL)
  if(thetaFlag["sinamp"]) {
    if(is.null(object@alpha1$sinamp$shape) || is.null(object@alpha1$sinamp$scale)) {
      stop("Prior parameters for TSMod1@alpha1$sinamp have not been set")
    }
    samples$sinamp <- rgamma(n, shape=object@alpha1$sinamp$shape, scale=object@alpha1$sinamp$scale)
  } else {
    samples$sinamp <- rep.int(times=n, x=thetaFixed$sinamp)
  }
  if(thetaFlag["freq"]) {
    if(is.null(object@alpha1$freq$shape) || is.null(object@alpha1$freq$scale)) {
      stop("Prior parameters for TSMod1@alpha1$freq have not been set")
    }
    samples$freq <- rgamma(n, shape=object@alpha1$freq$shape, scale=object@alpha1$freq$scale)
  } else {
    samples$freq <- rep.int(times=n, x=thetaFixed$freq)
  } 
  if(thetaFlag["phase"]) {
    samples$phase <- runif(n, min=0, max=1) 
  } else {
    samples$phase <- rep.int(times=n, x=thetaFixed$phase)
  }
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("FuncSinusoid"), definition=sample.prior.theta.FuncSinusoid)
#
trans.theta.FuncSinusoid <- function(object, thetaFlag) {
  if(thetaFlag["sinamp"]) {
    object@theta1$sinamp <- log10(object@theta1$sinamp)
  }
  if(thetaFlag["freq"]) {
    object@theta1$freq <- log10(object@theta1$freq)
  }
# v20: No phase transformation at all now that the prior is unity over 0-1 and zero outside.
#  if(thetaFlag["phase"]) { # No need to do anything if you use the circular boundary condition instead
#    # WARNING: if phase=1 this gets transformed to +Inf (due to divide by zero)
#    # WARNING: if phase=0 this gets transformed to -Inf 
#    object@theta1$phase <- log(object@theta1$phase/(1-object@theta1$phase)) # logit 
#  }
  return( object@theta1[names(new("FuncSinusoid")@theta1)] )
}
setMethod(trans.theta, signature=c("FuncSinusoid"), definition=trans.theta.FuncSinusoid)
#
invtrans.theta.FuncSinusoid <- function(object, thetaFlag) {
  if(thetaFlag["sinamp"]) {
    object@theta1$sinamp <- 10^object@theta1$sinamp
  }
  if(thetaFlag["freq"]) {
    object@theta1$freq <- 10^object@theta1$freq
  }
# v20: No phase transformation at all now that the prior is unity over 0-1 and zero outside.
#  if(thetaFlag["phase"]) {
#    object@theta1$phase <- 1/(1+exp(-object@theta1$phase)) # sigmoid (transforms +Inf and -Inf okay)
##    object@theta1$phase <- object@theta1$phase %% 1 # use this for the circular boundary condition
#  }
  return( object@theta1[names(new("FuncSinusoid")@theta1)] )
}
setMethod(invtrans.theta, signature=c("FuncSinusoid"), definition=invtrans.theta.FuncSinusoid)
#
jacobian.theta.FuncSinusoid <- function(object, thetaFlag) {
  jac <- 1
  if(thetaFlag["sinamp"]) {    
    jac <- jac * log(10) * exp(object@theta1$sinamp)
  }
  if(thetaFlag["freq"]) {
    jac <- jac * log(10) * exp(object@theta1$freq)
  }
  return(jac)
}
setMethod(jacobian.theta, signature=c("FuncSinusoid"), definition=jacobian.theta.FuncSinusoid)


#################### FuncSinusoidZero

# FuncSinusoidZero (sinamp, freq, phase)
# P(sinamp) ~ Gamma(shape, scale)  P(freq) ~ Gamma(shape, scale)  P(phase) ~ Uniform()
# As FuncSinusoid but symmetric in signal about zero (rather than having zero as the minimum)
setClass("FuncSinusoidZero", contains="FuncSinusoid")
#
eval.mod.FuncSinusoidZero <- function(object) {
  return( (object@theta1$sinamp/2) * cos(2*pi*(object@theta1$freq*object@t + object@theta1$phase)) )
}
setMethod(eval.mod, signature=c("FuncSinusoidZero"), definition=eval.mod.FuncSinusoidZero)
# No need to explicitly define prior evaluation/sampling or transformation/jacobian methods, as they are identical and inherited


#################### FuncQuasiSinusoid

# FuncQuasiSinusoid (sinamp, freq, phase, fp1, fp2, fp3, fp4)
# P(sinamp) ~ Gamma(shape, scale)  P(freq) ~ Gamma(shape, scale)  P(phase) ~ Uniform(0,1)
# P(fpX) ~ Uniform(-1,+1) X=1..Nfp
# The number of phase points, Nfp, is hard-wired into the class definition (in the prototype).
# Nfp must be 2 or more (the maximum no. turning points in fitted function is Np-1).
# fptimes is a numeric vector of size Nfp+1 specifying the times at which the phase points are
# specified. They must be provided in the intialization and then remain fixed.
# The "quasi sinusoidal" aspect is formed by fitting a cubic spline to fpX vs fptimes and using
# that to predict the phase shift, vphase, at any time.
# Note that fpX has range (-1,+1), but trans.theta.FuncQuasiSinusoid() prevents it getting very close
# to these boundaries in order to avoid numerical problems and thus transformation errors.
# This also has the side affect of initializations of the model parameters outside this range 
# apparently being accepted, because they are truncated to this range before being checked by 
# the MCMC algorithm in mcmc.R
setClass("FuncQuasiSinusoid", contains="TSMod1SC", representation=representation(Nfp="numeric", fptimes="numeric"), prototype=prototype(t=NULL, Nfp=4, fptimes=NULL, theta1=list(sinamp=NULL, freq=NULL, phase=NULL, fp1=NULL, fp2=NULL, fp3=NULL, fp4=NULL), alpha1=list(sinamp=list(shape=NULL, scale=NULL), freq=list(shape=NULL, scale=NULL), phase=NULL, fp1=NULL, fp2=NULL, fp3=NULL, fp4=NULL) ) )
#
init.FuncQuasiSinusoid <- function(.Object, fptimes=rep.int(0, .Object@Nfp+1)) {
  if(length(fptimes) != .Object@Nfp+1) stop("init.FuncQuasiSinusoid: fptimes must have length ", .Object@Nfp+1)
  .Object@fptimes <- fptimes
  .Object
}
setMethod("initialize", signature="FuncQuasiSinusoid", definition=init.FuncQuasiSinusoid)
#
eval.mod.FuncQuasiSinusoid <- function(object) {
  fpvec  <- c(0, object@theta1$fp1, object@theta1$fp2, object@theta1$fp3, object@theta1$fp4)
  vphase <- splint(object@fptimes, fpvec, xgrid=object@t) # df defaults to length(fpvec), i.e. lambda=0
  return( (object@theta1$sinamp/2) * ( cos(2*pi*(object@theta1$freq*object@t + object@theta1$phase + vphase)) + 1 ) )
}
setMethod(eval.mod, signature=c("FuncQuasiSinusoid"), definition=eval.mod.FuncQuasiSinusoid)
#
eval.prior.theta.FuncQuasiSinusoid <- function(object) {
  # Note: P(fpX) ~ Uniform()  implicitly set to 1 as it is independent of fpX
  sinampPrior <- ifelse( (is.null(object@alpha1$sinamp$shape) || is.null(object@alpha1$sinamp$scale)), 1, dgamma(object@theta1$sinamp, shape=object@alpha1$sinamp$shape, scale=object@alpha1$sinamp$scale) )
  freqPrior <- ifelse ( (is.null(object@alpha1$freq$shape) || is.null(object@alpha1$freq$scale)), 1, dgamma(object@theta1$freq, shape=object@alpha1$freq$shape, scale=object@alpha1$freq$scale) )
  phasePrior <- dunif(object@theta1$phase, min=0, max=1)
  return( prod(sinampPrior, freqPrior, phasePrior) )
}
setMethod(eval.prior.theta, signature=c("FuncQuasiSinusoid"), definition=eval.prior.theta.FuncQuasiSinusoid)
#
sample.prior.theta.FuncQuasiSinusoid <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(sinamp=NULL, freq=NULL, phase=NULL, fp1=NULL, fp2=NULL, fp3=NULL, fp4=NULL)
  if(thetaFlag["sinamp"]) {
    if(is.null(object@alpha1$sinamp$shape) || is.null(object@alpha1$sinamp$scale)) {
      stop("Prior parameters for TSMod1@alpha1$sinamp have not been set")
    }
    samples$sinamp <- rgamma(n, shape=object@alpha1$sinamp$shape, scale=object@alpha1$sinamp$scale)
  } else {
    samples$sinamp <- rep.int(times=n, x=thetaFixed$sinamp)
  }
  if(thetaFlag["freq"]) {
    if(is.null(object@alpha1$freq$shape) || is.null(object@alpha1$freq$scale)) {
      stop("Prior parameters for TSMod1@alpha1$freq have not been set")
    }
    samples$freq <- rgamma(n, shape=object@alpha1$freq$shape, scale=object@alpha1$freq$scale)
  } else {
    samples$freq <- rep.int(times=n, x=thetaFixed$freq)
  } 
  if(thetaFlag["phase"]) {
    samples$phase <- runif(n, min=0, max=1) 
  } else {
    samples$phase <- rep.int(times=n, x=thetaFixed$phase)
  }
  if(thetaFlag["fp1"]) {
    samples$fp1 <- runif(n, min=-1, max=1) 
  } else {
    samples$fp1 <- rep.int(times=n, x=thetaFixed$fp1)
  }
  if(thetaFlag["fp2"]) {
    samples$fp2 <- runif(n, min=-1, max=1) 
  } else {
    samples$fp2 <- rep.int(times=n, x=thetaFixed$fp2)
  }
  if(thetaFlag["fp3"]) {
    samples$fp3 <- runif(n, min=-1, max=1) 
  } else {
    samples$fp3 <- rep.int(times=n, x=thetaFixed$fp3)
  }
  if(thetaFlag["fp4"]) {
    samples$fp4 <- runif(n, min=-1, max=1) 
  } else {
    samples$fp4 <- rep.int(times=n, x=thetaFixed$fp4)
  }
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("FuncQuasiSinusoid"), definition=sample.prior.theta.FuncQuasiSinusoid)
#
trans.theta.FuncQuasiSinusoid <- function(object, thetaFlag) {
  translim <- 1e-10 # this is used to keep fpX within the range -1+translim to 1-translim
                    # in order to avoid the warnings listed below for the transformation
                    # which would otherwise occur due to finite machine precision.
                    # If it exceeds either limit it is trucated to the limit.
  if(thetaFlag["sinamp"]) {
    object@theta1$sinamp <- log10(object@theta1$sinamp)
  }
  if(thetaFlag["freq"]) {
    object@theta1$freq <- log10(object@theta1$freq)
  }
# v20: No phase transformation at all now that the prior is unity over 0-1 and zero outside.
#  if(thetaFlag["phase"]) { # No need to do anything if you use the circular boundary condition instead
#   # WARNING: if phase=1 this gets transformed to +Inf (due to divide by zero)
#   # WARNING: if phase=0 this gets transformed to -Inf 
#    object@theta1$phase <- log(object@theta1$phase/(1-object@theta1$phase)) # logit (inverse is the sigmoid)
#  } 
  if(thetaFlag["fp1"]) {
    # WARNING: if fp1=+1 this gets transformed to +Inf (due to divide by zero)
    # WARNING: if fp1=-1 this gets transformed to -Inf
    # This applies also to others below. But it is now avoided by use of translim to truncate fpX
    x <- ifelse(1-abs(object@theta1$fp1) > translim, object@theta1$fp1, sign(object@theta1$fp1)*(1-translim))
    object@theta1$fp1 <- log((1+x)/(1-x)) # = 2atanh(x), which is as logit but for (-1,+1)
  } 
  if(thetaFlag["fp2"]) { 
    x <- ifelse(1-abs(object@theta1$fp2) > translim, object@theta1$fp2, sign(object@theta1$fp2)*(1-translim))
    object@theta1$fp2 <- log((1+x)/(1-x))
  } 
  if(thetaFlag["fp3"]) { 
    x <- ifelse(1-abs(object@theta1$fp3) > translim, object@theta1$fp3, sign(object@theta1$fp3)*(1-translim))
    object@theta1$fp3 <- log((1+x)/(1-x))
  } 
  if(thetaFlag["fp4"]) { 
    x <- ifelse(1-abs(object@theta1$fp4) > translim, object@theta1$fp4, sign(object@theta1$fp4)*(1-translim))
    object@theta1$fp4 <- log((1+x)/(1-x))
  } 
  return( object@theta1[names(new("FuncQuasiSinusoid")@theta1)] )
}
setMethod(trans.theta, signature=c("FuncQuasiSinusoid"), definition=trans.theta.FuncQuasiSinusoid)
#
invtrans.theta.FuncQuasiSinusoid <- function(object, thetaFlag) {
  invtranslim <- -23.719 # the transformed fpX is kept above this value (by truncating up smaller values)
                         # in order to avoid the warnings listed below for the inverse transformation,
                         # which would otherwise occur due to finite machine precision.
                         # This value corresponds to the transformed value of fpX=-1+translim, translim=1e-10
  if(thetaFlag["sinamp"]) {
    object@theta1$sinamp <- 10^object@theta1$sinamp
  }
  if(thetaFlag["freq"]) {
    object@theta1$freq <- 10^object@theta1$freq
  }
# v20: No phase transformation at all now that the prior is unity over 0-1 and zero outside.
#  if(thetaFlag["phase"]) {
##    object@theta1$phase <- 1/(1+exp(-object@theta1$phase)) # sigmoid (transforms +Inf and -Inf okay)
#    object@theta1$phase <- object@theta1$phase %% 1 # use this for the circular boundary condition
#  }
  if(thetaFlag["fp1"]) {
    # WARNING: if fp=-Inf this gets transformed to NaN (+Inf is okay: is transformed to +1)
    # This applies also to others below. But it is now avoided by use of invtranslim to truncate fpX
    y <- exp(-1*ifelse(object@theta1$fp1 > invtranslim, object@theta1$fp1, invtranslim))
    object@theta1$fp1 <- (1-y)/(1+y)
  }
  if(thetaFlag["fp2"]) {
    y <- exp(-1*ifelse(object@theta1$fp2 > invtranslim, object@theta1$fp2, invtranslim))
    object@theta1$fp2 <- (1-y)/(1+y)
  }
  if(thetaFlag["fp3"]) {
    y <- exp(-1*ifelse(object@theta1$fp3 > invtranslim, object@theta1$fp3, invtranslim))
    object@theta1$fp3 <- (1-y)/(1+y)
  }
  if(thetaFlag["fp4"]) {
    y <- exp(-1*ifelse(object@theta1$fp4 > invtranslim, object@theta1$fp4, invtranslim))
    object@theta1$fp4 <- (1-y)/(1+y)
  }
  return( object@theta1[names(new("FuncQuasiSinusoid")@theta1)] )
}
setMethod(invtrans.theta, signature=c("FuncQuasiSinusoid"), definition=invtrans.theta.FuncQuasiSinusoid)
#
jacobian.theta.FuncQuasiSinusoid <- function(object, thetaFlag) {
  jac <- 1
  if(thetaFlag["sinamp"]) {
    jac <- jac * log(10) * exp(object@theta1$sinamp)
  }
  if(thetaFlag["freq"]) {
    jac <- jac * log(10) * exp(object@theta1$freq)
  }
  if(thetaFlag["fp1"]) {
    y   <- exp(-1*object@theta1$fp1)
    jac <- jac*y/(1+y)^2
  }
  if(thetaFlag["fp2"]) {
    y   <- exp(-1*object@theta1$fp2)
    jac <- jac*y/(1+y)^2
  }
  if(thetaFlag["fp3"]) {
    y   <- exp(-1*object@theta1$fp3)
    jac <- jac*y/(1+y)^2
  }
  if(thetaFlag["fp4"]) {
    y   <- exp(-1*object@theta1$fp4)
    jac <- jac*y/(1+y)^2
  }
  return(jac)
}
setMethod(jacobian.theta, signature=c("FuncQuasiSinusoid"), definition=jacobian.theta.FuncQuasiSinusoid)


#################### FuncQuasiSinusoidZero

# FuncQuasiSinusoid (sinamp, freq, phase, fp1, fp2, fp3, fp4)
# P(sinamp) ~ Gamma(shape, scale)  P(freq) ~ Gamma(shape, scale)  P(phase) ~ Uniform()
# P(fpX) ~ Uniform() X=1..Nfp
# As FuncQuasiSinusoid but symmetric in signal about zero (rather than having zero as the minimum)
setClass("FuncQuasiSinusoidZero", contains="FuncQuasiSinusoid")
#
eval.mod.FuncQuasiSinusoidZero <- function(object) {
  fpvec  <- c(0, object@theta1$fp1, object@theta1$fp2, object@theta1$fp3, object@theta1$fp4)
  vphase <- splint(object@fptimes, fpvec, xgrid=object@t) # df defaults to length(fpvec), i.e. lambda=0
  return( (object@theta1$sinamp/2) * cos(2*pi*(object@theta1$freq*object@t + object@theta1$phase + vphase)) )
}
setMethod(eval.mod, signature=c("FuncQuasiSinusoidZero"), definition=eval.mod.FuncQuasiSinusoidZero)
# No need to explicitly define prior evaluation/sampling or transformation/jacobian methods, as they are identical and inherited


#################### FuncLinear

# FuncLinear (slope, tmid, offset)
# slope: P(slope) ~ Uniform() in atan(slope), an angle, with constraint parameter enumerated type "sign":
#   sign="pos"    => slope >=0
#   sign="neg"    => slope <=0
#   otherwise no constraint
#   NOTE: This model is unusual in a number of senses concerning the slope parameter:
#   (1) the prior PDF is not on the parameter, but on a transformation of it, namely atan(slope)
#   (2) the prior PDF is uniform with no numerical parameters (alpha1), but a flag ("sign") indicating the range
#   (3) if the limit on slope is imposed, the evaluation of the prior can be zero.
#       This is not a problem, but it will throw a warning in calc.post.mcmc().
# tmid: P(tmid) ~ N(mean, sd)
# offset: Choice of prior selected by specifying priorform, e.g. in new("FuncUniform", priorform="Gamma")
#   P(offset) ~ Gamma(shape, scale)  when priorform="Gamma" 
#   P(offset) ~ Gaussian(mean, sd)   when priorform="Gaussian" [default]
#   This inherits the class FuncUniform for the offset parameter
setClass("FuncLinear", contains="FuncUniform", prototype=prototype(t=NULL, theta1=list(slope=NULL, tmid=NULL, offset=NULL), alpha1=list(slope=list(sign=NULL), tmid=list(mean=NULL, sd=NULL), offset=list(mean=NULL, sd=NULL)) ) )  ##### I suspect we can use offset=NULL for alpha1
#
eval.mod.FuncLinear <- function(object) {
  return( object@theta1$slope*(object@t - object@theta1$tmid) + eval.mod.FuncUniform(object) )
}
setMethod(eval.mod, signature=c("FuncLinear"), definition=eval.mod.FuncLinear)
#
eval.prior.theta.FuncLinear <- function(object) {
  tmidPrior <- ifelse( (is.null(object@alpha1$tmid$mean) || is.null(object@alpha1$tmid$sd)), 1, dnorm(object@theta1$tmid, mean=object@alpha1$tmid$mean, sd=object@alpha1$tmid$sd) )
  if( is.null(object@alpha1$slope$sign) ) {
    slopePrior <- 1
  } else {
    s <- object@theta1$slope
    slopePrior <- dcauchy(s, location=0, scale=1) # if object@alpha1$slope$sign is anything other than "pos" or "neg"
    if(object@alpha1$slope$sign=="pos") {
      slopePrior <- ifelse(s>=0, 2*slopePrior, 0)
    }
    if(object@alpha1$slope$sign=="neg") {
      slopePrior <- ifelse(s<=0, 2*slopePrior, 0)
    }
  }
  return( prod(tmidPrior, slopePrior, eval.prior.theta.FuncUniform(object)) )
}
setMethod(eval.prior.theta, signature=c("FuncLinear"), definition=eval.prior.theta.FuncLinear)
#
sample.prior.theta.FuncLinear <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(slope=NULL, tmid=NULL, offset=NULL)
  if(thetaFlag["slope"]) {
    if(is.null(object@alpha1$slope$sign)) {
      stop("Prior parameters for TSMod1@alpha1$slope have not been set")
    }
    samples$slope <- rcauchy(n, location=0, scale=1)
    if(object@alpha1$slope$sign=="pos") {samples$slope <-    abs(samples$slope)}
    if(object@alpha1$slope$sign=="neg") {samples$slope <- -1*abs(samples$slope)}
  } else {
    samples$slope <- rep.int(times=n, x=thetaFixed$slope)
  }
  if(thetaFlag["tmid"]) {
    if(is.null(object@alpha1$tmid$mean) || is.null(object@alpha1$tmid$sd)) {
      stop("Prior parameters for TSMod1@alpha1$tmid have not been set")
    }
    samples$tmid <- rnorm(n, mean=object@alpha1$tmid$mean, sd=object@alpha1$tmid$sd)
  } else {
    samples$tmid <- rep.int(times=n, x=thetaFixed$tmid)
  }
  samples$offset <- sample.prior.theta.FuncUniform(object, thetaFlag, thetaFixed, n)$offset
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("FuncLinear"), definition=sample.prior.theta.FuncLinear)
#
trans.theta.FuncLinear <- function(object, thetaFlag) {
  if(thetaFlag["slope"]) { # sampler works on angle in radians (-pi/2 to +pi/2): circular boundary condition
    object@theta1$slope <- atan(object@theta1$slope) 
  }
  object@theta1$offset <- trans.theta.FuncUniform(object, thetaFlag)$offset
  return( object@theta1[names(new("FuncLinear")@theta1)] )
}
setMethod(trans.theta, signature=c("FuncLinear"), definition=trans.theta.FuncLinear)
#
invtrans.theta.FuncLinear <- function(object, thetaFlag) {
  if(thetaFlag["slope"]) { 
    object@theta1$slope <- tan(object@theta1$slope) 
  }
  object@theta1$offset <- invtrans.theta.FuncUniform(object, thetaFlag)$offset
  return( object@theta1[names(new("FuncLinear")@theta1)] )
}
setMethod(invtrans.theta, signature=c("FuncLinear"), definition=invtrans.theta.FuncLinear)
#
jacobian.theta.FuncLinear <- function(object, thetaFlag) {
  jac <- jacobian.theta.FuncUniform(object, thetaFlag)
  if(thetaFlag["slope"]) {
    jac <- jac * 1/(cos(object@theta1$slope))^2
  }
  return(jac)
}
setMethod(jacobian.theta, signature=c("FuncLinear"), definition=jacobian.theta.FuncLinear)


#################### FuncCombUniformSigmoid
 
# FuncCombUniformSigmoid (offset, sigamp, lambda, tzero)
# Combines additively FuncUniform and FuncSigmoid (using class inheritance)
setClass("FuncCombUniformSigmoid", contains=c("FuncUniform", "FuncSigmoid"), prototype=prototype(t=NULL, theta1=list(offset=NULL, sigamp=NULL, lambda=NULL, tzero=NULL), alpha1=list(offset=list(shape=NULL, scale=NULL), sigamp=list(shape=NULL, scale=NULL), lambda=list(mean=NULL, sd=NULL), tzero=list(mean=NULL, sd=NULL)) ) ) 
#
eval.mod.FuncCombUniformSigmoid <- function(object) {
  return( eval.mod.FuncUniform(object) + eval.mod.FuncSigmoid(object) )
}
setMethod(eval.mod, signature=c("FuncCombUniformSigmoid"), definition=eval.mod.FuncCombUniformSigmoid)
#
eval.prior.theta.FuncCombUniformSigmoid <- function(object) {
  return( eval.prior.theta.FuncUniform(object) * eval.prior.theta.FuncSigmoid(object) )
}
setMethod(eval.prior.theta, signature=c("FuncCombUniformSigmoid"), definition=eval.prior.theta.FuncCombUniformSigmoid)
#
sample.prior.theta.FuncCombUniformSigmoid <- function(object, thetaFlag, thetaFixed, n) {
  return( cbind(sample.prior.theta.FuncUniform(object, thetaFlag, thetaFixed, n), sample.prior.theta.FuncSigmoid(object, thetaFlag, thetaFixed, n)) )
}
setMethod(sample.prior.theta, signature=c("FuncCombUniformSigmoid"), definition=sample.prior.theta.FuncCombUniformSigmoid)
#
trans.theta.FuncCombUniformSigmoid <- function(object, thetaFlag) {
  return( c(trans.theta.FuncUniform(object, thetaFlag), trans.theta.FuncSigmoid(object, thetaFlag)) )
}
setMethod(trans.theta, signature=c("FuncCombUniformSigmoid"), definition=trans.theta.FuncCombUniformSigmoid)
#
invtrans.theta.FuncCombUniformSigmoid <- function(object, thetaFlag) {
  return( c(invtrans.theta.FuncUniform(object, thetaFlag), invtrans.theta.FuncSigmoid(object, thetaFlag)) )
}
setMethod(invtrans.theta, signature=c("FuncCombUniformSigmoid"), definition=invtrans.theta.FuncCombUniformSigmoid)
#
jacobian.theta.FuncCombUniformSigmoid <- function(object, thetaFlag) {
  return( prod(jacobian.theta.FuncUniform(object, thetaFlag), jacobian.theta.FuncSigmoid(object, thetaFlag)) )
}
setMethod(jacobian.theta, signature=c("FuncCombUniformSigmoid"), definition=jacobian.theta.FuncCombUniformSigmoid)


#################### FuncCombUniformSigmoidZero 

# FuncCombUniformSigmoidZero (offset, sigamp, lambda, tzero)
# Combines additively FuncUniform and FuncSigmoidZero
# (using class inheritance from FuncCombUniformSigmoid rather than multiple inheritance from
# FuncUniform and FuncSigmoidZero, to reduce the code repetition)
setClass("FuncCombUniformSigmoidZero", contains="FuncCombUniformSigmoid")
#
eval.mod.FuncCombUniformSigmoidZero <- function(object) {
   return( eval.mod.FuncUniform(object) + eval.mod.FuncSigmoidZero(object) )
}
setMethod(eval.mod, signature=c("FuncCombUniformSigmoidZero"), definition=eval.mod.FuncCombUniformSigmoidZero)
# No need to explicitly define prior evaluation/sampling and transformation/jacobian methods, as they are identical and inherited


#################### FuncCombUniformSinusoid

# FuncCombUniformSinusoid (offset, sinamp, freq, phase)
# Combines additively FuncUniform and FuncSinusoid (using class inheritance)
setClass("FuncCombUniformSinusoid", contains=c("FuncUniform", "FuncSinusoid"), prototype=prototype(t=NULL, theta1=list(offset=NULL, sinamp=NULL, freq=NULL, phase=NULL), alpha1=list(offset=list(shape=NULL, scale=NULL), sinamp=list(shape=NULL, scale=NULL), freq=list(shape=NULL, scale=NULL), phase=NULL) ) )
#
eval.mod.FuncCombUniformSinusoid <- function(object) {
  return( eval.mod.FuncUniform(object) + eval.mod.FuncSinusoid(object) )
}
setMethod(eval.mod, signature=c("FuncCombUniformSinusoid"), definition=eval.mod.FuncCombUniformSinusoid)
#
eval.prior.theta.FuncCombUniformSinusoid <- function(object) {
  return( eval.prior.theta.FuncUniform(object) * eval.prior.theta.FuncSinusoid(object) )
}
setMethod(eval.prior.theta, signature=c("FuncCombUniformSinusoid"), definition=eval.prior.theta.FuncCombUniformSinusoid)
#
sample.prior.theta.FuncCombUniformSinusoid <- function(object, thetaFlag, thetaFixed, n) {
  return( cbind(sample.prior.theta.FuncUniform(object, thetaFlag, thetaFixed, n), sample.prior.theta.FuncSinusoid(object, thetaFlag, thetaFixed, n)) )
}
setMethod(sample.prior.theta, signature=c("FuncCombUniformSinusoid"), definition=sample.prior.theta.FuncCombUniformSinusoid)
#
trans.theta.FuncCombUniformSinusoid <- function(object, thetaFlag) {
  return( c(trans.theta.FuncUniform(object, thetaFlag), trans.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(trans.theta, signature=c("FuncCombUniformSinusoid"), definition=trans.theta.FuncCombUniformSinusoid)
#
invtrans.theta.FuncCombUniformSinusoid <- function(object, thetaFlag) {
  return( c(invtrans.theta.FuncUniform(object, thetaFlag), invtrans.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(invtrans.theta, signature=c("FuncCombUniformSinusoid"), definition=invtrans.theta.FuncCombUniformSinusoid)
#
jacobian.theta.FuncCombUniformSinusoid <- function(object, thetaFlag) {
  return( prod(jacobian.theta.FuncUniform(object, thetaFlag), jacobian.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(jacobian.theta, signature=c("FuncCombUniformSinusoid"), definition=jacobian.theta.FuncCombUniformSinusoid)


#################### FuncCombUniformSinusoidZero

# FuncCombUniformSinusoidZero  (offset, sinamp, freq, phase)
# Combines additively FuncUniform and FuncSinusoidZero
# (using class inheritance from FuncCombUniformSinusoid rather than multiple inheritance from
# FuncUniform and FuncSinusoidZero, to reduce the code repetition)
setClass("FuncCombUniformSinusoidZero", contains="FuncCombUniformSinusoid")
#
eval.mod.FuncCombUniformSinusoidZero <- function(object) {
   return( eval.mod.FuncUniform(object) + eval.mod.FuncSinusoidZero(object) )
}
setMethod(eval.mod, signature=c("FuncCombUniformSinusoidZero"), definition=eval.mod.FuncCombUniformSinusoidZero)
# No need to explicitly define prior evaluation/sampling or transformation/jacobian methods, as they are identical and inherited


#################### FuncCombUniformQuasiSinusoid

# FuncCombUniformQuasiSinusoid (offset, sinamp, freq, phase, fp1, fp2, fp3, fp4)
# Combines additively FuncUniform and FuncQuasiSinusoid (using class inheritance)
setClass("FuncCombUniformQuasiSinusoid", contains=c("FuncUniform", "FuncQuasiSinusoid"), prototype=prototype(t=NULL, Nfp=4, fptimes=NULL, theta1=list(offset=NULL, sinamp=NULL, freq=NULL, phase=NULL, fp1=NULL, fp2=NULL, fp3=NULL, fp4=NULL), alpha1=list(offset=list(shape=NULL, scale=NULL), sinamp=list(shape=NULL, scale=NULL), freq=list(shape=NULL, scale=NULL), phase=NULL, fp1=NULL, fp2=NULL, fp3=NULL, fp4=NULL) ) )
#
# As we're using multiple inheritance where there is more than one initializer involved, we need to call
# them explicitly with all the required arguments (and no others). This might not be the "proper" way to do it.
# Note that we do need the defaults in the definition.
init.FuncCombUniformQuasiSinusoid <- function(.Object, priorform="Gaussian", fptimes=rep.int(0, .Object@Nfp+1)) {
  .Object <- init.FuncUniform(.Object, priorform)
  .Object <- init.FuncQuasiSinusoid(.Object, fptimes)
  .Object
}
setMethod("initialize", signature="FuncCombUniformQuasiSinusoid", definition=init.FuncCombUniformQuasiSinusoid)
#
eval.mod.FuncCombUniformQuasiSinusoid <- function(object) {
  return( eval.mod.FuncUniform(object) + eval.mod.FuncQuasiSinusoid(object) )
}
setMethod(eval.mod, signature=c("FuncCombUniformQuasiSinusoid"), definition=eval.mod.FuncCombUniformQuasiSinusoid)
#
eval.prior.theta.FuncCombUniformQuasiSinusoid <- function(object) {
  return( eval.prior.theta.FuncUniform(object) * eval.prior.theta.FuncQuasiSinusoid(object) )
}
setMethod(eval.prior.theta, signature=c("FuncCombUniformQuasiSinusoid"), definition=eval.prior.theta.FuncCombUniformQuasiSinusoid)
#
sample.prior.theta.FuncCombUniformQuasiSinusoid <- function(object, thetaFlag, thetaFixed, n) {
  return( cbind(sample.prior.theta.FuncUniform(object, thetaFlag, thetaFixed, n), sample.prior.theta.FuncQuasiSinusoid(object, thetaFlag, thetaFixed, n)) )
}
setMethod(sample.prior.theta, signature=c("FuncCombUniformQuasiSinusoid"), definition=sample.prior.theta.FuncCombUniformQuasiSinusoid)
#
trans.theta.FuncCombUniformQuasiSinusoid <- function(object, thetaFlag) {
  return( c(trans.theta.FuncUniform(object, thetaFlag), trans.theta.FuncQuasiSinusoid(object, thetaFlag)) )
}
setMethod(trans.theta, signature=c("FuncCombUniformQuasiSinusoid"), definition=trans.theta.FuncCombUniformQuasiSinusoid)
#
invtrans.theta.FuncCombUniformQuasiSinusoid <- function(object, thetaFlag) {
  return( c(invtrans.theta.FuncUniform(object, thetaFlag), invtrans.theta.FuncQuasiSinusoid(object, thetaFlag)) )
}
setMethod(invtrans.theta, signature=c("FuncCombUniformQuasiSinusoid"), definition=invtrans.theta.FuncCombUniformQuasiSinusoid)
#
jacobian.theta.FuncCombUniformQuasiSinusoid <- function(object, thetaFlag) {
  return( prod(jacobian.theta.FuncUniform(object, thetaFlag), jacobian.theta.FuncQuasiSinusoid(object, thetaFlag)) )
}
setMethod(jacobian.theta, signature=c("FuncCombUniformQuasiSinusoid"), definition=jacobian.theta.FuncCombUniformQuasiSinusoid)


#################### FuncCombUniformQuasiSinusoidZero

# FuncCombUniformQuasiSinusoidZero  (offset, sinamp, freq, phase, fp1, fp2, fp3, fp4)
# Combines additively FuncUniform and FuncQuasiSinusoidZero
# (using class inheritance from FuncCombUniformQuasiSinusoid rather than multiple inheritance from
# FuncUniform and FuncSinusoidZero, to reduce the code repetition)
setClass("FuncCombUniformQuasiSinusoidZero", contains="FuncCombUniformQuasiSinusoid")
#
eval.mod.FuncCombUniformQuasiSinusoidZero <- function(object) {
   return( eval.mod.FuncUniform(object) + eval.mod.FuncQuasiSinusoidZero(object) )
}
setMethod(eval.mod, signature=c("FuncCombUniformQuasiSinusoidZero"), definition=eval.mod.FuncCombUniformQuasiSinusoidZero)
# No need to explicitly define prior evaluation/sampling or transformation/jacobian methods, as they are identical and inherited


#################### FuncCombUniformSigmoidSinusoid

# FuncCombUniformSigmoidSinusoid (offset, sigamp, lambda, tzero, sinamp, freq, phase)
# Combines additively FuncUniform, FuncSigmoid and FuncSinusoid (using class inheritance)
setClass("FuncCombUniformSigmoidSinusoid", contains=c("FuncUniform", "FuncSigmoid", "FuncSinusoid"), prototype=prototype(t=NULL, theta1=list(offset=NULL, sigamp=NULL, lambda=NULL, tzero=NULL, sinamp=NULL, freq=NULL, phase=NULL), alpha1=list(offset=list(shape=NULL, scale=NULL), sigamp=list(shape=NULL, scale=NULL), lambda=list(mean=NULL, sd=NULL), tzero=list(mean=NULL, sd=NULL), sinamp=list(shape=NULL, scale=NULL), freq=list(shape=NULL, scale=NULL), phase=NULL) ) ) 
#
eval.mod.FuncCombUniformSigmoidSinusoid <- function(object) {
  return( eval.mod.FuncUniform(object) + eval.mod.FuncSigmoid(object) + eval.mod.FuncSinusoid(object)  )
}
setMethod(eval.mod, signature=c("FuncCombUniformSigmoidSinusoid"), definition=eval.mod.FuncCombUniformSigmoidSinusoid)
#
eval.prior.theta.FuncCombUniformSigmoidSinusoid <- function(object) {
  return( eval.prior.theta.FuncUniform(object) * eval.prior.theta.FuncSigmoid(object) * eval.prior.theta.FuncSinusoid(object) )
}
setMethod(eval.prior.theta, signature=c("FuncCombUniformSigmoidSinusoid"), definition=eval.prior.theta.FuncCombUniformSigmoidSinusoid)
#
sample.prior.theta.FuncCombUniformSigmoidSinusoid <- function(object, thetaFlag, thetaFixed, n) {
  return( cbind(sample.prior.theta.FuncUniform(object, thetaFlag, thetaFixed, n), sample.prior.theta.FuncSigmoid(object, thetaFlag, thetaFixed, n), sample.prior.theta.FuncSinusoid(object, thetaFlag, thetaFixed, n)) )
}
setMethod(sample.prior.theta, signature=c("FuncCombUniformSigmoidSinusoid"), definition=sample.prior.theta.FuncCombUniformSigmoidSinusoid)
#
trans.theta.FuncCombUniformSigmoidSinusoid <- function(object, thetaFlag) {
  return( c(trans.theta.FuncUniform(object, thetaFlag), trans.theta.FuncSigmoid(object, thetaFlag), trans.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(trans.theta, signature=c("FuncCombUniformSigmoidSinusoid"), definition=trans.theta.FuncCombUniformSigmoidSinusoid)
#
invtrans.theta.FuncCombUniformSigmoidSinusoid <- function(object, thetaFlag) {
  return( c(invtrans.theta.FuncUniform(object, thetaFlag), invtrans.theta.FuncSigmoid(object, thetaFlag), invtrans.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(invtrans.theta, signature=c("FuncCombUniformSigmoidSinusoid"), definition=invtrans.theta.FuncCombUniformSigmoidSinusoid)
#
jacobian.theta.FuncCombUniformSigmoidSinusoid <- function(object, thetaFlag) {
  return( prod(jacobian.theta.FuncUniform(object, thetaFlag), jacobian.theta.FuncSigmoid(object, thetaFlag), jacobian.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(jacobian.theta, signature=c("FuncCombUniformSigmoidSinusoid"), definition=jacobian.theta.FuncCombUniformSigmoidSinusoid)


#################### FuncCombUniformSigmoidZeroSinusoidZero

# FuncCombUniformSigmoidZeroSinusoidZero (offset, sigamp, lambda, tzero, sinamp, freq, phase)
# Combines additively FuncUniform, FuncSigmoidZero and FuncSinusoidZero (using class inheritance)
setClass("FuncCombUniformSigmoidZeroSinusoidZero", contains=c("FuncUniform", "FuncSigmoidZero", "FuncSinusoidZero"), prototype=prototype(t=NULL, theta1=list(offset=NULL, sigamp=NULL, lambda=NULL, tzero=NULL, sinamp=NULL, freq=NULL, phase=NULL), alpha1=list(offset=list(shape=NULL, scale=NULL), sigamp=list(shape=NULL, scale=NULL), lambda=list(mean=NULL, sd=NULL), tzero=list(mean=NULL, sd=NULL), sinamp=list(shape=NULL, scale=NULL), freq=list(shape=NULL, scale=NULL), phase=NULL) ) ) 
#
eval.mod.FuncCombUniformSigmoidZeroSinusoidZero <- function(object) {
  return( eval.mod.FuncUniform(object) + eval.mod.FuncSigmoidZero(object) + eval.mod.FuncSinusoidZero(object)  )
}
setMethod(eval.mod, signature=c("FuncCombUniformSigmoidZeroSinusoidZero"), definition=eval.mod.FuncCombUniformSigmoidZeroSinusoidZero)
#
# v15: The methods eval.prior.theta, sample.prior.theta, trans.theta, invtrans.theta use the "non-Zero" functions (methods)
# in their definition because these are identical to the "Zero" versions. That's why explicit methods for things like
# eval.prior.theta.FuncSigmoidZero() do not exist - we use eval.prior.theta.FuncSigmoid() instead.
eval.prior.theta.FuncCombUniformSigmoidZeroSinusoidZero <- function(object) {
  return( eval.prior.theta.FuncUniform(object) * eval.prior.theta.FuncSigmoid(object) * eval.prior.theta.FuncSinusoid(object) )
}
setMethod(eval.prior.theta, signature=c("FuncCombUniformSigmoidZeroSinusoidZero"), definition=eval.prior.theta.FuncCombUniformSigmoidZeroSinusoidZero)
#
sample.prior.theta.FuncCombUniformSigmoidZeroSinusoidZero <- function(object, thetaFlag, thetaFixed, n) {
  return( cbind(sample.prior.theta.FuncUniform(object, thetaFlag, thetaFixed, n), sample.prior.theta.FuncSigmoid(object, thetaFlag, thetaFixed, n), sample.prior.theta.FuncSinusoid(object, thetaFlag, thetaFixed, n)) )
}
setMethod(sample.prior.theta, signature=c("FuncCombUniformSigmoidZeroSinusoidZero"), definition=sample.prior.theta.FuncCombUniformSigmoidZeroSinusoidZero)
#
trans.theta.FuncCombUniformSigmoidZeroSinusoidZero <- function(object, thetaFlag) {
  return( c(trans.theta.FuncUniform(object, thetaFlag), trans.theta.FuncSigmoid(object, thetaFlag), trans.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(trans.theta, signature=c("FuncCombUniformSigmoidZeroSinusoidZero"), definition=trans.theta.FuncCombUniformSigmoidZeroSinusoidZero)
#
invtrans.theta.FuncCombUniformSigmoidZeroSinusoidZero <- function(object, thetaFlag) {
  return( c(invtrans.theta.FuncUniform(object, thetaFlag), invtrans.theta.FuncSigmoid(object, thetaFlag), invtrans.theta.FuncSinusoid(object, thetaFlag)) )
}
setMethod(invtrans.theta, signature=c("FuncCombUniformSigmoidZeroSinusoidZero"), definition=invtrans.theta.FuncCombUniformSigmoidZeroSinusoidZero)
#
jacobian.theta.FuncCombUniformSigmoidZeroSinusoidZero <- function(object, thetaFlag) {
  return( prod(jacobian.theta.FuncUniform(object, thetaFlag), jacobian.theta.FuncSigmoidZero(object, thetaFlag), jacobian.theta.FuncSinusoidZero(object, thetaFlag)) )
}
setMethod(jacobian.theta, signature=c("FuncCombUniformSigmoidZeroSinusoidZero"), definition=jacobian.theta.FuncCombUniformSigmoidZeroSinusoidZero)


#################### BypassTSMod1

# BypassTSMod1
# Methods are defined by the virtual class BypassSC
setClass("BypassTSMod1", contains="BypassSC", representation=representation(theta1="list", alpha1="list"), prototype=prototype(theta1=list(dummy=NULL) ) )



#################### TSMod2. TS model, first stochastic part.
#################### Returns P(z|t,theta1,theta2,TSMod1,TSMod2) if using TSMod1 to represent the mean of this function.
#################### For full stochastic models returns either P(z|t,theta2,TSMod2) or parameters of PDF at each event.

# superclass
setClass("TSMod2SC", representation=representation(z="numeric", theta2="list", TSMod1SC="ANY", alpha2="list"))


#################### ProbGaussian

# ProbGaussian (sd)
# P(sd) ~ Gamma(shape, scale)
setClass("ProbGaussian", contains="TSMod2SC", prototype=prototype(z=NULL, theta2=list(sd=NULL), TSMod1SC=NULL, alpha2=list(sd=list(shape=NULL, scale=NULL)) ) )
#
# Return scalar, or vector of length(object@z)
eval.mod.ProbGaussian <- function(object) {
  return( dnorm(object@z, mean=eval.mod(object@TSMod1SC), sd=object@theta2$sd, log=FALSE) )
}
setMethod(eval.mod, signature=c("ProbGaussian"), definition=eval.mod.ProbGaussian)
#
eval.prior.theta.ProbGaussian <- function(object) {
  return( ifelse( (is.null(object@alpha2$sd$shape) || is.null(object@alpha2$sd$scale)), 1, dgamma(object@theta2$sd, shape=object@alpha2$sd$shape, scale=object@alpha2$sd$scale) ) )
}
setMethod(eval.prior.theta, signature=c("ProbGaussian"), definition=eval.prior.theta.ProbGaussian)
#
sample.prior.theta.ProbGaussian <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(sd=NULL)
  if(thetaFlag["sd"]) {
    if(is.null(object@alpha2$sd$shape) || is.null(object@alpha2$sd$scale)) {
      stop("Prior parameters for TSMod2@alpha2$sd have not been set")
    }
    samples$sd <- rgamma(n, shape=object@alpha2$sd$shape, scale=object@alpha2$sd$scale)
  } else {
    samples$sd <- rep.int(times=n, x=thetaFixed$sd)
  }
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("ProbGaussian"), definition=sample.prior.theta.ProbGaussian)
#
trans.theta.ProbGaussian <- function(object, thetaFlag) {
  if(thetaFlag["sd"]) {
    object@theta2$sd <- log10(object@theta2$sd)
  }
  return( object@theta2[names(new("ProbGaussian")@theta2)] )
}
setMethod(trans.theta, signature=c("ProbGaussian"), definition=trans.theta.ProbGaussian)
#
invtrans.theta.ProbGaussian <- function(object, thetaFlag) {
  if(thetaFlag["sd"]) {
    object@theta2$sd <- 10^object@theta2$sd
  }
  return( object@theta2[names(new("ProbGaussian")@theta2)] )
}
setMethod(invtrans.theta, signature=c("ProbGaussian"), definition=invtrans.theta.ProbGaussian)
#
jacobian.theta.ProbGaussian <- function(object, thetaFlag) {
  if(thetaFlag["sd"]) {
    return( log(10) * exp(object@theta2$sd) )
  } else {
    return(1)  
  }
}
setMethod(jacobian.theta, signature=c("ProbGaussian"), definition=jacobian.theta.ProbGaussian)


#################### ProbOUprocess

# ProbOUprocess (relax, diffcon, offset, startmean, startsd)
# The OU process is a stochastic 1D process, z(t).
# It is stationary, Markov and Gaussian with time-variable mean and variance.
# (One example process is the velocity of a particle undergoing Brownian motion.)
# Parameters and their prior PDF forms:
# relax       relaxation time (tau)     P(relax)      ~ Gamma(shape, scale)   
# diffcon     diffusion constant (c)    P(diffcon)    ~ Gamma(shape, scale)
# offset      long-term mean (b)        P(offset)     ~ N(mean, sd)
# zstartmean  mean of prior of z(t=s1)  P(zstartmean) ~ N(mean, sd)
# zstartsd    sd   of prior of z(t=s1)  P(zstartsd)   ~ Gamma(shape, scale)
# s1 = measured time of the first event.
# Note that zstartmean is the absolute value. It is not relative to offset.
# TSMod1 should be bypassed.
setClass("ProbOUprocess", contains="TSMod2SC", prototype=prototype(t=NULL, z=NULL, theta2=list(relax=NULL, diffcon=NULL, offset=NULL, zstartmean=NULL, zstartsd=NULL), alpha2=list(relax=list(shape=NULL, scale=NULL), diffcon=list(shape=NULL, scale=NULL), offset=list(mean=NULL, sd=NULL), zstartmean=list(mean=NULL, sd=NULL), zstartsd=list(shape=NULL, scale=NULL)) ) )
#
# Return a ncol(obsdata) x 2 dataframe with named columns zMean and zVar which give the mean and variance
# of the "prior" PDF of the OU process variable at each event (see TN)
eval.mod.ProbOUprocess <- function(object, obsdata) {
  Nevents  <- nrow(obsdata)
  zMean    <- vector(mode="numeric", length=Nevents)
  zVar     <- vector(mode="numeric", length=Nevents)
  j <- 1
  zMean[j] <- object@theta2$zstartmean
  zVar[j]  <- object@theta2$zstartsd^2
  for(j in 1:Nevents) {
    if(j > 1) {
      # WARNING: divide by zero if relax=0 (controlled by prior)
      nu <- exp(-(obsdata[j,1]-obsdata[j-1,1])/object@theta2$relax)
      Vz <- (object@theta2$diffcon*object@theta2$relax/2)*(1-nu^2)
      zMean[j] <- zMeanPost*nu + object@theta2$offset*(1-nu)
      zVar[j]  <- zVarPost*nu^2 + Vz
    }
    if(j < Nevents) { # calculate posterior
      ysigsq    <- obsdata[j,4]^2
      denom     <- ysigsq + zVar[j]
      # WARNING: divide by zero if denom=0 (controlled by data/prior)
      zMeanPost <- (obsdata[j,3]*zVar[j] + zMean[j]*ysigsq)/denom
      zVarPost  <- zVar[j]*ysigsq/denom
    }
  }
  return(data.frame(zMean=zMean, zVar=zVar))
}  
setMethod(eval.mod, signature=c("ProbOUprocess"), definition=eval.mod.ProbOUprocess)
#
eval.prior.theta.ProbOUprocess <- function(object) {
  relaxPrior <- ifelse( (is.null(object@alpha2$relax$shape) || is.null(object@alpha2$relax$scale)), 1, dgamma(object@theta2$relax, shape=object@alpha2$relax$shape, scale=object@alpha2$relax$scale) )
  diffconPrior <- ifelse( (is.null(object@alpha2$diffcon$shape) || is.null(object@alpha2$diffcon$scale)), 1, dgamma(object@theta2$diffcon, shape=object@alpha2$diffcon$shape, scale=object@alpha2$diffcon$scale) )
  offsetPrior <- ifelse( (is.null(object@alpha2$offset$mean) || is.null(object@alpha2$offset$sd)), 1, dnorm(object@theta2$offset, mean=object@alpha2$offset$mean, sd=object@alpha2$offset$sd) )
  zstartmeanPrior <- ifelse( (is.null(object@alpha2$zstartmean$mean) || is.null(object@alpha2$zstartmean$sd)), 1, dnorm(object@theta2$zstartmean, mean=object@alpha2$zstartmean$mean, sd=object@alpha2$zstartmean$sd) )
  zstartsdPrior <- ifelse( (is.null(object@alpha2$zstartsd$shape) || is.null(object@alpha2$zstartsd$scale)), 1, dgamma(object@theta2$zstartsd, shape=object@alpha2$zstartsd$shape, scale=object@alpha2$zstartsd$scale) )
  return( prod(relaxPrior, diffconPrior, offsetPrior, zstartmeanPrior, zstartsdPrior) )
}
setMethod(eval.prior.theta, signature=c("ProbOUprocess"), definition=eval.prior.theta.ProbOUprocess)
#
sample.prior.theta.ProbOUprocess <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(relax=NULL, diffcon=NULL, offset=NULL, zstartmean=NULL, zstartsd=NULL)

  if(thetaFlag["relax"]) {
    if(is.null(object@alpha2$relax$shape) || is.null(object@alpha2$relax$scale)) {
      stop("Prior parameters for TSMod2@alpha2$relax have not been set")
    }
    samples$relax <- rgamma(n, shape=object@alpha2$relax$shape, scale=object@alpha2$relax$scale)
  } else {
    samples$relax <- rep.int(times=n, x=thetaFixed$relax)
  }

  if(thetaFlag["diffcon"]) {
    if(is.null(object@alpha2$diffcon$shape) || is.null(object@alpha2$diffcon$scale)) {
      stop("Prior parameters for TSMod2@alpha2$diffcon have not been set")
    }
    samples$diffcon <- rgamma(n, shape=object@alpha2$diffcon$shape, scale=object@alpha2$diffcon$scale)
  } else {
    samples$diffcon <- rep.int(times=n, x=thetaFixed$diffcon)
  } 

  if(thetaFlag["offset"]) {
    if(is.null(object@alpha2$offset$mean) || is.null(object@alpha2$offset$sd)) {
      stop("Prior parameters for TSMod2@alpha2$offset have not been set")
    }
    samples$offset <- rnorm(n, mean=object@alpha2$offset$mean, sd=object@alpha2$offset$sd) 
  } else {
    samples$offset <- rep.int(times=n, x=thetaFixed$offset)
  }

  if(thetaFlag["zstartmean"]) {
    if(is.null(object@alpha2$zstartmean$mean) || is.null(object@alpha2$zstartmean$sd)) {
      stop("Prior parameters for TSMod2@alpha2$zstartmean have not been set")
    }
    samples$zstartmean <- rnorm(n, mean=object@alpha2$zstartmean$mean, sd=object@alpha2$zstartmean$sd) 
  } else {
    samples$zstartmean <- rep.int(times=n, x=thetaFixed$zstartmean)
  }

  if(thetaFlag["zstartsd"]) {
    if(is.null(object@alpha2$zstartsd$shape) || is.null(object@alpha2$zstartsd$scale)) {
      stop("Prior parameters for TSMod2@alpha2$zstartsd have not been set")
    }
    samples$zstartsd <- rgamma(n, shape=object@alpha2$zstartsd$shape, scale=object@alpha2$zstartsd$scale)
  } else {
    samples$zstartsd <- rep.int(times=n, x=thetaFixed$zstartsd)
  } 

  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("ProbOUprocess"), definition=sample.prior.theta.ProbOUprocess)
#
trans.theta.ProbOUprocess <- function(object, thetaFlag) {
  if(thetaFlag["relax"]) {
    object@theta2$relax <- log10(object@theta2$relax)
  }
  if(thetaFlag["diffcon"]) {
    object@theta2$diffcon <- log10(object@theta2$diffcon)
  }
  if(thetaFlag["zstartsd"]) {
    object@theta2$zstartsd <- log10(object@theta2$zstartsd)
  }
  return( object@theta2[names(new("ProbOUprocess")@theta2)] )
}
setMethod(trans.theta, signature=c("ProbOUprocess"), definition=trans.theta.ProbOUprocess)
#
invtrans.theta.ProbOUprocess <- function(object, thetaFlag) {
  if(thetaFlag["relax"]) {
    object@theta2$relax <- 10^(object@theta2$relax)
  }
  if(thetaFlag["diffcon"]) {
    object@theta2$diffcon <- 10^(object@theta2$diffcon)
  }
  if(thetaFlag["zstartsd"]) {
    object@theta2$zstartsd <- 10^(object@theta2$zstartsd)
  }
  return( object@theta2[names(new("ProbOUprocess")@theta2)] )
}
setMethod(invtrans.theta, signature=c("ProbOUprocess"), definition=invtrans.theta.ProbOUprocess)
#
jacobian.theta.ProbOUprocess <- function(object, thetaFlag) {
  jac <- 1
  if(thetaFlag["relax"]) {    
    jac <- jac * log(10) * exp(object@theta2$relax)
  }
  if(thetaFlag["diffcon"]) {    
    jac <- jac * log(10) * exp(object@theta2$diffcon)
  }
  if(thetaFlag["zstartsd"]) {    
    jac <- jac * log(10) * exp(object@theta2$zstartsd)
  }
  return(jac)
}
setMethod(jacobian.theta, signature=c("ProbOUprocess"), definition=jacobian.theta.ProbOUprocess)


#################### ProbWienerprocess (3 August 2012)

# ProbWienerprocess (diffcon, offset, startmean, startsd)
# The Wiener process is a stochastic 1D process, z(t).
# This model is based on the OU process one using different mean and variance.
#   The only differences are that the relax parameter has been removed, and
#   and the expressions for Vz, zMean[j] and zVar[j] are different.
# Parameters and their prior PDF forms:
# diffcon     diffusion constant (c)    P(diffcon)    ~ Gamma(shape, scale)
# offset      long-term mean (b)        P(offset)     ~ N(mean, sd)
# zstartmean  mean of prior of z(t=s1)  P(zstartmean) ~ N(mean, sd)
# zstartsd    sd   of prior of z(t=s1)  P(zstartsd)   ~ Gamma(shape, scale)
# s1 = measured time of the first event.
# Note that zstartmean is the absolute value. It is not relative to offset.
# TSMod1 should be bypassed.
setClass("ProbWienerprocess", contains="TSMod2SC", prototype=prototype(t=NULL, z=NULL, theta2=list(diffcon=NULL, offset=NULL, zstartmean=NULL, zstartsd=NULL), alpha2=list(diffcon=list(shape=NULL, scale=NULL), offset=list(mean=NULL, sd=NULL), zstartmean=list(mean=NULL, sd=NULL), zstartsd=list(shape=NULL, scale=NULL)) ) )
#
# Return a ncol(obsdata) x 2 dataframe with named columns zMean and zVar which give the mean and variance
# of the "prior" PDF of the Wiener process variable at each event (see TN)
eval.mod.ProbWienerprocess <- function(object, obsdata) {
  Nevents  <- nrow(obsdata)
  zMean    <- vector(mode="numeric", length=Nevents)
  zVar     <- vector(mode="numeric", length=Nevents)
  j <- 1
  zMean[j] <- object@theta2$zstartmean
  zVar[j]  <- object@theta2$zstartsd^2
  for(j in 1:Nevents) {
    if(j > 1) {
      Dt <- obsdata[j,1]-obsdata[j-1,1]
      Vz <- object@theta2$diffcon*Dt
      zMean[j] <- zMeanPost + object@theta2$offset
      zVar[j]  <- zVarPost  + Vz
    }
    if(j < Nevents) { # calculate posterior
      ysigsq    <- obsdata[j,4]^2
      denom     <- ysigsq + zVar[j]
      # WARNING: divide by zero if denom=0 (controlled by data/prior)
      zMeanPost <- (obsdata[j,3]*zVar[j] + zMean[j]*ysigsq)/denom
      zVarPost  <- zVar[j]*ysigsq/denom
    }
  }
  return(data.frame(zMean=zMean, zVar=zVar))
}  
setMethod(eval.mod, signature=c("ProbWienerprocess"), definition=eval.mod.ProbWienerprocess)
#
eval.prior.theta.ProbWienerprocess <- function(object) {
  diffconPrior <- ifelse( (is.null(object@alpha2$diffcon$shape) || is.null(object@alpha2$diffcon$scale)), 1, dgamma(object@theta2$diffcon, shape=object@alpha2$diffcon$shape, scale=object@alpha2$diffcon$scale) )
  offsetPrior <- ifelse( (is.null(object@alpha2$offset$mean) || is.null(object@alpha2$offset$sd)), 1, dnorm(object@theta2$offset, mean=object@alpha2$offset$mean, sd=object@alpha2$offset$sd) )
  zstartmeanPrior <- ifelse( (is.null(object@alpha2$zstartmean$mean) || is.null(object@alpha2$zstartmean$sd)), 1, dnorm(object@theta2$zstartmean, mean=object@alpha2$zstartmean$mean, sd=object@alpha2$zstartmean$sd) )
  zstartsdPrior <- ifelse( (is.null(object@alpha2$zstartsd$shape) || is.null(object@alpha2$zstartsd$scale)), 1, dgamma(object@theta2$zstartsd, shape=object@alpha2$zstartsd$shape, scale=object@alpha2$zstartsd$scale) )
  return( prod(diffconPrior, offsetPrior, zstartmeanPrior, zstartsdPrior) )
}
setMethod(eval.prior.theta, signature=c("ProbWienerprocess"), definition=eval.prior.theta.ProbWienerprocess)
#
sample.prior.theta.ProbWienerprocess <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(diffcon=NULL, offset=NULL, zstartmean=NULL, zstartsd=NULL)

  if(thetaFlag["diffcon"]) {
    if(is.null(object@alpha2$diffcon$shape) || is.null(object@alpha2$diffcon$scale)) {
      stop("Prior parameters for TSMod2@alpha2$diffcon have not been set")
    }
    samples$diffcon <- rgamma(n, shape=object@alpha2$diffcon$shape, scale=object@alpha2$diffcon$scale)
  } else {
    samples$diffcon <- rep.int(times=n, x=thetaFixed$diffcon)
  } 

  if(thetaFlag["offset"]) {
    if(is.null(object@alpha2$offset$mean) || is.null(object@alpha2$offset$sd)) {
      stop("Prior parameters for TSMod2@alpha2$offset have not been set")
    }
    samples$offset <- rnorm(n, mean=object@alpha2$offset$mean, sd=object@alpha2$offset$sd) 
  } else {
    samples$offset <- rep.int(times=n, x=thetaFixed$offset)
  }

  if(thetaFlag["zstartmean"]) {
    if(is.null(object@alpha2$zstartmean$mean) || is.null(object@alpha2$zstartmean$sd)) {
      stop("Prior parameters for TSMod2@alpha2$zstartmean have not been set")
    }
    samples$zstartmean <- rnorm(n, mean=object@alpha2$zstartmean$mean, sd=object@alpha2$zstartmean$sd) 
  } else {
    samples$zstartmean <- rep.int(times=n, x=thetaFixed$zstartmean)
  }

  if(thetaFlag["zstartsd"]) {
    if(is.null(object@alpha2$zstartsd$shape) || is.null(object@alpha2$zstartsd$scale)) {
      stop("Prior parameters for TSMod2@alpha2$zstartsd have not been set")
    }
    samples$zstartsd <- rgamma(n, shape=object@alpha2$zstartsd$shape, scale=object@alpha2$zstartsd$scale)
  } else {
    samples$zstartsd <- rep.int(times=n, x=thetaFixed$zstartsd)
  } 

  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("ProbWienerprocess"), definition=sample.prior.theta.ProbWienerprocess)
#
trans.theta.ProbWienerprocess <- function(object, thetaFlag) {
  if(thetaFlag["diffcon"]) {
    object@theta2$diffcon <- log10(object@theta2$diffcon)
  }
  if(thetaFlag["zstartsd"]) {
    object@theta2$zstartsd <- log10(object@theta2$zstartsd)
  }
  return( object@theta2[names(new("ProbWienerprocess")@theta2)] )
}
setMethod(trans.theta, signature=c("ProbWienerprocess"), definition=trans.theta.ProbWienerprocess)
#
invtrans.theta.ProbWienerprocess <- function(object, thetaFlag) {
  if(thetaFlag["diffcon"]) {
    object@theta2$diffcon <- 10^(object@theta2$diffcon)
  }
  if(thetaFlag["zstartsd"]) {
    object@theta2$zstartsd <- 10^(object@theta2$zstartsd)
  }
  return( object@theta2[names(new("ProbWienerprocess")@theta2)] )
}
setMethod(invtrans.theta, signature=c("ProbWienerprocess"), definition=invtrans.theta.ProbWienerprocess)
#
jacobian.theta.ProbWienerprocess <- function(object, thetaFlag) {
  jac <- 1
  if(thetaFlag["diffcon"]) {    
    jac <- jac * log(10) * exp(object@theta2$diffcon)
  }
  if(thetaFlag["zstartsd"]) {    
    jac <- jac * log(10) * exp(object@theta2$zstartsd)
  }
  return(jac)
}
setMethod(jacobian.theta, signature=c("ProbWienerprocess"), definition=jacobian.theta.ProbWienerprocess)


#################### ProbOUprocNaive

# ProbOUprocNaive (relax, diffcon)
# This not only extends TSMod2SC but also adds news slot (specified in representation).
# This is the OU process. See Gillespie (1996), Phys. Rev. E, vol. 54, p. 2084 for details.
# In terms of Brownian motion, x is velocity and w is position (Gillepsie calls this "y").
# This class describes x, which has a Gaussian PDF with a time-variable mean and variance.
# Parameters (in parentheses in notation of Gillespie) and their prior PDF forms:
# relax     relaxation time (tau)     P(relax)   ~ Gamma(shape, scale)   
# diffcon   diffusion constant (c)    P(diffcon) ~ Gamma(shape, scale)
# tzero     (t0) - non-adjustable
# xzero     (x0)                      P(xzero)   ~ N(mean, sd)
# tzero is fixed and no attempt should be made to adjust it.
# For compatibility purposes the prior PDF is unity for this parameters (which is invisible here)
# and the prior sampling returns its fixed value. TSMod1 should be bypassed. 
#
setClass("ProbOUprocNaive", contains="TSMod2SC", representation=representation(t="numeric"), prototype=prototype(t=NULL, z=NULL, theta2=list(relax=NULL, diffcon=NULL, tzero=NULL, xzero=NULL), alpha2=list(relax=list(shape=NULL, scale=NULL), diffcon=list(shape=NULL, scale=NULL), tzero=NA, xzero=list(mean=NULL, sd=NULL)) ) )
#
# Return scalar, or vector of length(object@t)
eval.mod.ProbOUprocNaive <- function(object) {
  # WARNING: divide by zero if relax=0 (controlled by data/prior)
  ndt   <- (object@t - object@theta2$tzero)/object@theta2$relax
  meanX <- object@theta2$xzero*exp(-ndt)
  varX  <- (1/2)*object@theta2$diffcon*object@theta2$relax*(1-exp(-2*ndt))
  if(varX<0) {
    stop("Negative varX in eval.mod.ProbOUprocNaive, probably due to object@t < object@theta2$tzero")
  }
  return( dnorm(object@z, mean=meanX, sd=sqrt(varX), log=FALSE) )
}
setMethod(eval.mod, signature=c("ProbOUprocNaive"), definition=eval.mod.ProbOUprocNaive)
#
eval.prior.theta.ProbOUprocNaive <- function(object) {
  relaxPrior <- ifelse( (is.null(object@alpha2$relax$shape) || is.null(object@alpha2$relax$scale)), 1, dgamma(object@theta2$relax, shape=object@alpha2$relax$shape, scale=object@alpha2$relax$scale) )
  diffconPrior <- ifelse( (is.null(object@alpha2$diffcon$shape) || is.null(object@alpha2$diffcon$scale)), 1, dgamma(object@theta2$diffcon, shape=object@alpha2$diffcon$shape, scale=object@alpha2$diffcon$scale) )
  xzeroPrior <- ifelse( (is.null(object@alpha2$xzero$mean) || is.null(object@alpha2$xzero$sd)), 1, dnorm(object@theta2$xzero, mean=object@alpha2$xzero$mean, sd=object@alpha2$xzero$sd) )
  return( prod(relaxPrior, diffconPrior, xzeroPrior) )
}
setMethod(eval.prior.theta, signature=c("ProbOUprocNaive"), definition=eval.prior.theta.ProbOUprocNaive)
#
sample.prior.theta.ProbOUprocNaive <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(relax=NULL, diffcon=NULL, tzero=NULL, xzero=NULL)

  if(thetaFlag["relax"]) {
    if(is.null(object@alpha2$relax$shape) || is.null(object@alpha2$relax$scale)) {
      stop("Prior parameters for TSMod2@alpha2$relax have not been set")
    }
    samples$relax <- rgamma(n, shape=object@alpha2$relax$shape, scale=object@alpha2$relax$scale)
  } else {
    samples$relax <- rep.int(times=n, x=thetaFixed$relax)
  }

  if(thetaFlag["diffcon"]) {
    if(is.null(object@alpha2$diffcon$shape) || is.null(object@alpha2$diffcon$scale)) {
      stop("Prior parameters for TSMod2@alpha2$diffcon have not been set")
    }
    samples$diffcon <- rgamma(n, shape=object@alpha2$diffcon$shape, scale=object@alpha2$diffcon$scale)
  } else {
    samples$diffcon <- rep.int(times=n, x=thetaFixed$diffcon)
  } 

  samples$tzero <- rep.int(times=n, x=thetaFixed$tzero)

  if(thetaFlag["xzero"]) {
    if(is.null(object@alpha2$xzero$mean) || is.null(object@alpha2$xzero$sd)) {
      stop("Prior parameters for TSMod2@alpha2$xzero have not been set")
    }
    samples$xzero <- rnorm(n, mean=object@alpha2$xzero$mean, sd=object@alpha2$xzero$sd) 
  } else {
    samples$xzero <- rep.int(times=n, x=thetaFixed$xzero)
  }

  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("ProbOUprocNaive"), definition=sample.prior.theta.ProbOUprocNaive)


#################### ProbIntOUprocNaive

# ProbIntOUprocNaive
# The integrated OU process, i.e. describes the variable w (rather than x)
# as a Gaussian PDF with a time-variable mean and variance.
# This class is like ProbOUprocNaive but with different functional forms for mean and
# variance, and with an extra parameter:
# wzero   P(wzero) ~ N(mean, sd)
#
setClass("ProbIntOUprocNaive", contains="ProbOUprocNaive", representation=representation(t="numeric"), prototype=prototype(t=NULL, z=NULL, theta2=list(relax=NULL, diffcon=NULL, tzero=NULL, xzero=NULL, wzero=NULL), alpha2=list(relax=list(shape=NULL, scale=NULL), diffcon=list(shape=NULL, scale=NULL), tzero=NA, xzero=list(mean=NULL, sd=NULL), wzero=list(mean=NULL, sd=NULL)) ) )
#
# Return scalar, or vector of length(object@t)
eval.mod.ProbIntOUprocNaive <- function(object) {
  muc   <- object@theta2$diffcon * object@theta2$relax^3
  ndt   <- (object@t - object@theta2$tzero)/object@theta2$relax
  meanW <- object@theta2$wzero + object@theta2$xzero*object@theta2$relax*(1-exp(-ndt))
  varW  <- muc * ( ndt - 2*(1-exp(-ndt)) + (1/2)*(1-exp(-2*ndt)) )
  if(varW<0) {
    stop("Negative varW in eval.mod.ProbIntOUprocNaive, probably due to object@t < object@theta2$tzero")
  }
  return( dnorm(object@z, mean=meanW, sd=sqrt(varW), log=FALSE) )
}
setMethod(eval.mod, signature=c("ProbIntOUprocNaive"), definition=eval.mod.ProbIntOUprocNaive)
#
eval.prior.theta.ProbIntOUprocNaive <- function(object) {
  wzeroPrior <- ifelse( (is.null(object@alpha2$wzero$mean) || is.null(object@alpha2$wzero$sd)), 1, dnorm(object@theta2$wzero, mean=object@alpha2$wzero$mean, sd=object@alpha2$wzero$sd) )
  return( eval.prior.theta.ProbOUprocNaive(object) * wzeroPrior )
}
setMethod(eval.prior.theta, signature=c("ProbIntOUprocNaive"), definition=eval.prior.theta.ProbIntOUprocNaive)
#
sample.prior.theta.ProbIntOUprocNaive <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(wzero=NULL)
  if(thetaFlag["wzero"]) {
    if(is.null(object@alpha2$wzero$mean) || is.null(object@alpha2$wzero$sd)) {
      stop("Prior parameters for TSMod2@alpha2$wzero have not been set")
    }
    samples$wzero <- rnorm(n, mean=object@alpha2$wzero$mean, sd=object@alpha2$wzero$sd) 
  } else {
    samples$wzero <- rep.int(times=n, x=thetaFixed$wzero)
  }

  return( cbind(sample.prior.theta.ProbOUprocNaive(object, thetaFlag, thetaFixed, n), as.data.frame(samples)) )
}
setMethod(sample.prior.theta, signature=c("ProbIntOUprocNaive"), definition=sample.prior.theta.ProbIntOUprocNaive)


# BypassTSMod2
# Methods are defined by the virtual class BypassSC
setClass("BypassTSMod2", contains="BypassSC", representation=representation(theta2="list", alpha2="list"), prototype=prototype(theta2=list(dummy=NULL) ) )



####################
#################### TSMod3. TS model, second stochastic part. Returns P(t|theta3,TSMod3) = f(t, theta3 ; TSMod3)
####################

# superclass
setClass("TSMod3SC", representation=representation(t="numeric", theta3="list", alpha3="list"))


#################### ProbUniform

# ProbUniform (tmin, tmax)
# No prior: theta parameters are fixed => no attempt should be made to adjust this parameter
# For compatibility with a generic TSMod3:
# eval.prior.theta() returns unity, sample.prior.theta() returns the fixed values.
#
setClass("ProbUniform", contains="TSMod3SC", prototype=prototype(t=NULL, theta3=list(tmin=NULL, tmax=NULL), alpha3=list(tmin=NA, tmax=NA)) )
#
# Return scalar, or vector of length(object@t)
eval.mod.ProbUniform <- function(object) {
    return( dunif(object@t, min=object@theta3$tmin, max=object@theta3$tmax, log=FALSE) )
}
setMethod(eval.mod, signature=c("ProbUniform"), definition=eval.mod.ProbUniform)
#
eval.prior.theta.ProbUniform <- function(object) { 
  return(1)
}
setMethod(eval.prior.theta, signature=c("ProbUniform"), definition=eval.prior.theta.ProbUniform)
#
sample.prior.theta.ProbUniform <- function(object, thetaFlag, thetaFixed, n) { 
  samples <- list(tmin=NULL, tmax=NULL)
  samples$tmin <- rep.int(times=n, x=thetaFixed$tmin)
  samples$tmax <- rep.int(times=n, x=thetaFixed$tmax)
  return(as.data.frame(samples))
}
setMethod(sample.prior.theta, signature=c("ProbUniform"), definition=sample.prior.theta.ProbUniform)
#
trans.theta.ProbUniform <- function(object) {
    return( object@theta3[names(new("ProbUniform")@theta3)] ) # Parameter always fixed: no transform required. Could make this default method.
}
setMethod(trans.theta, signature=c("ProbUniform"), definition=trans.theta.ProbUniform)
#
invtrans.theta.ProbUniform <- function(object) {
    return( object@theta3[names(new("ProbUniform")@theta3)] ) # Parameter always fixed: no transform required. Could make this default method.
}
setMethod(invtrans.theta, signature=c("ProbUniform"), definition=invtrans.theta.ProbUniform)



#################### MeasMod. Measurement model. Returns P(D_j | t_j, z_j, sigma) where 
#################### D_j = (s, y) and sigma = (s.sd, y.sd)
#################### obsdata = c(s, s.sd, y, y.sd) [Accessed by position, not name]

# superclass
setClass("MeasModSC", representation=representation(t="numeric", z="numeric", obsdata="numeric"))


#################### TwoGaussian

# TwoGaussian
# both time and signal 1D Gaussians. Likelihood is their product
setClass("TwoGaussian", contains="MeasModSC", prototype=prototype(t=NULL, z=NULL, obsdata=NULL) )
#
# Return scalar, or vector of longer of length(object@t) and length(object@z)
# (Behaviour is not desirable if both are vectors of different lengths!)
eval.mod.TwoGaussian <- function(object) {
  return( dnorm(object@obsdata[1], mean=object@t, sd=object@obsdata[2]) * dnorm(object@obsdata[3], mean=object@z, sd=object@obsdata[4]) )
}
setMethod(eval.mod, signature=c("TwoGaussian"), definition=eval.mod.TwoGaussian)


#################### TimeUniformSignalGaussian

# TimeUniformSignalGaussian
# time (s) is Uniform, signal (y) is Gaussian. Likelihood is their product.
# s.sd is interpreted as sd of uniform distribution centered on t
setClass("TimeUniformSignalGaussian", contains="MeasModSC", prototype=prototype(t=NULL, z=NULL, obsdata=NULL) )
#
eval.mod.TimeUniformSignalGaussian <- function(object) {
  s.width <- sqrt(12)*object@obsdata[2]
  s.min   <- object@t - s.width/2
  s.max   <- object@t + s.width/2 
  t.like  <- dunif(object@obsdata[1], min=s.min, max=s.max)
  z.like  <- dnorm(object@obsdata[3], mean=object@z, sd=object@obsdata[4])
  return( t.like*z.like )
}
setMethod(eval.mod, signature=c("TimeUniformSignalGaussian"), definition=eval.mod.TimeUniformSignalGaussian)
